import io
from pathlib import Path

from lzallright import LZOCompressor

from unblob.file_utils import (
    Endian,
    FileSystem,
    InvalidInputFormat,
    StructParser,
)
from unblob.handlers.compression._ucl import UCLDecompressor
from unblob.models import (
    Extractor,
    ExtractResult,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

QNX_C_DEFINITION = """
    struct filehdr {
        char     signature[8];
        int32    usize;     // Uncompressed size of the file
        uint16   blksize;   // Size of compression blocks
        uint8    cmptype;   // Type of compression (CMP_LZO, ...)
        uint8    flags;
    } filehdr_t;

    struct cmphdr {
        uint16   prev;      // Offset to previous hdr
        uint16   next;      // Offset to next hdr
        uint16   pusize;    // Size of prev uncompressed blk
        uint16   usize;     // Size of this uncompressed blk
    } cmphdr_t;

    struct cmpblk {
	struct cmphdr	hdr;
	uint8_t			buf[32*1024];
    } cmpblk_t;
"""
SUPPORTED_BLOCK_SIZES = [2**i for i in range(12, 16)]


class QNXDeflateExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(QNX_C_DEFINITION)

    def extract(self, inpath: Path, outdir: Path) -> ExtractResult:
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            fhdr = self._struct_parser.cparser_le.filehdr_t(file)
            decompressor = LZOCompressor() if fhdr.cmptype == 0 else UCLDecompressor()
            with fs.open(Path(f"{inpath.stem}.uncompressed")) as outfile:
                cmphdr = self._struct_parser.cparser_le.cmphdr_t(file)
                while cmphdr.next != 0:
                    compressed_part = file.read(cmphdr.next - 8)
                    outfile.write(decompressor.decompress(compressed_part))
                    cmphdr = self._struct_parser.cparser_le.cmphdr_t(file)
            return ExtractResult(reports=fs.problems)


class QNXDeflateHandler(StructHandler):
    NAME = "qnx_deflate"
    C_DEFINITIONS = QNX_C_DEFINITION
    HEADER_STRUCT = "filehdr_t"
    PATTERNS = [
        HexString("69 77 6c 79 66 6d 62 70 [6] (00|01)")  # iwlyfmbp
    ]
    EXTRACTOR = QNXDeflateExtractor()
    DOC = HandlerDoc(
        name="QNX Deflate",
        description="QNX deflate are compressed files using a block-based compression format with either LZO or UCL algorithm.",
        handler_type=HandlerType.COMPRESSION,
        vendor="Blackberry",
        references=[
            Reference(title="QNX wikipedia", url="https://en.wikipedia.org/wiki/QNX"),
        ],
        limitations=[],
    )

    def is_valid_header(self, header) -> bool:
        return header.usize > 0 and header.blksize in SUPPORTED_BLOCK_SIZES

    def calculate_chunk(self, file, start_offset) -> ValidChunk:
        header = self.parse_header(file, Endian.LITTLE)
        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid QNX header")

        cmphdr = self.cparser_le.cmphdr_t(file)
        while cmphdr.next != 0:
            # return from 8 bytes because it parse the cmpheader
            file.seek(cmphdr.next - 8, io.SEEK_CUR)
            cmphdr = self.cparser_le.cmphdr_t(file)
        end_offset = file.tell()

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



import os
from threading import Lock

__last_id = 0
__last_id_lock = Lock()


def new_id():
    # NOTE, that uuid4 can not be used, as there are multiple processes at run time,
    # and as subprocesses inherit the random number state, so uuid4 would generate colliding ids
    # another option that would not work is to use uuid1, but we could generate the id
    # at the same time

    global __last_id  # noqa: PLW0603
    with __last_id_lock:
        __last_id += 1
        return f"{os.getpid()}:{__last_id}"



import lzma
import re
import zlib
from collections.abc import Callable
from pathlib import Path

import pyzstd

from unblob.file_utils import (
    Endian,
    FileSystem,
    InvalidInputFormat,
    StructParser,
    iterate_file,
)
from unblob.models import (
    Extractor,
    ExtractResult,
    File,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

# [Ref] https://github.com/freebsd/freebsd-src/tree/master/sys/geom/uzip
C_DEFINITIONS = r"""
    typedef struct uzip_header{
        char magic[16];
        char format[112];
        uint32_t block_size;
        uint32_t block_count;
        uint64_t toc[block_count];
    } uzip_header_t;
"""

HEADER_STRUCT = "uzip_header_t"

ZLIB_COMPRESSION = "#!/bin/sh\x0a#V2.0\x20"
LZMA_COMPRESSION = "#!/bin/sh\x0a#L3.0\x0a"
ZSTD_COMPRESSION = "#!/bin/sh\x0a#Z4.0\x20"


class Decompressor:
    DECOMPRESSOR: Callable

    def __init__(self):
        self._decompressor = self.DECOMPRESSOR()

    def decompress(self, data: bytes) -> bytes:
        return self._decompressor.decompress(data)

    def flush(self) -> bytes:
        return b""


class LZMADecompressor(Decompressor):
    DECOMPRESSOR = lzma.LZMADecompressor


class ZLIBDecompressor(Decompressor):
    DECOMPRESSOR = zlib.decompressobj

    def flush(self) -> bytes:
        return self._decompressor.flush()


class ZSTDDecompressor(Decompressor):
    DECOMPRESSOR = pyzstd.EndlessZstdDecompressor


DECOMPRESS_METHOD: dict[bytes, type[Decompressor]] = {
    ZLIB_COMPRESSION.encode(): ZLIBDecompressor,
    LZMA_COMPRESSION.encode(): LZMADecompressor,
    ZSTD_COMPRESSION.encode(): ZSTDDecompressor,
}


class UZIPExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        with File.from_path(inpath) as infile:
            parser = StructParser(C_DEFINITIONS)
            header = parser.parse(HEADER_STRUCT, infile, Endian.BIG)
            fs = FileSystem(outdir)
            outpath = Path(inpath.stem)

            try:
                decompressor_cls = DECOMPRESS_METHOD[header.magic]
            except LookupError:
                raise InvalidInputFormat("unsupported compression format") from None

            with fs.open(outpath, "wb+") as outfile:
                for current_offset, next_offset in zip(
                    header.toc[:-1], header.toc[1:], strict=False
                ):
                    compressed_len = next_offset - current_offset
                    if compressed_len == 0:
                        continue
                    decompressor = decompressor_cls()
                    for chunk in iterate_file(infile, current_offset, compressed_len):
                        outfile.write(decompressor.decompress(chunk))
                    outfile.write(decompressor.flush())
            return ExtractResult(reports=fs.problems)


class UZIPHandler(StructHandler):
    NAME = "uzip"
    PATTERNS = [
        Regex(re.escape(ZLIB_COMPRESSION)),
        Regex(re.escape(LZMA_COMPRESSION)),
        Regex(re.escape(ZSTD_COMPRESSION)),
    ]
    HEADER_STRUCT = HEADER_STRUCT
    C_DEFINITIONS = C_DEFINITIONS
    EXTRACTOR = UZIPExtractor()

    DOC = HandlerDoc(
        name="UZIP",
        description="FreeBSD UZIP is a block-based compressed disk image format. It uses a table of contents to index compressed blocks, supporting ZLIB, LZMA, and ZSTD compression algorithms.",
        handler_type=HandlerType.COMPRESSION,
        vendor="FreeBSD",
        references=[
            Reference(
                title="FreeBSD UZIP Documentation",
                url="https://github.com/freebsd/freebsd-src/tree/master/sys/geom/uzip",
            ),
        ],
        limitations=[],
    )

    def is_valid_header(self, header) -> bool:
        return (
            header.block_count > 0
            and header.block_size > 0
            and header.block_size % 512 == 0
        )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, Endian.BIG)

        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid uzip header.")

        # take the last TOC block offset, end of file is that block offset,
        # starting from the start offset
        end_offset = start_offset + header.toc[-1]
        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



"""Unix compress'ed chunk identification.

We identify the end offset of any identified unix compress'ed chunk by
performing Lempel-Ziv-Welch decompression on a chunk starting from the
identified start offset, and ending at the end of the whole file being
analyzed.

If we reach an invalid code or the stream ends in the middle of a
code, we do not recursively call the decompression with -1 size,
rather just fail on the chunk as we have seen too many false-positives
picked up by this heuristic.

Once the decompression procedure works without errors, that means we
have a valid chunk and can return its current end offset.

We use a small heuristic to return the right end offset.  This
heuristic tends to work well when arbitrary data appended at the end
of the stream is made of random bytes (no repeating letters, no large
set of ASCII letters).

It obviously can be wrong from time to time, leading to a compress'ed
chunk that we can decompress (obviously), but uncompressed data will
contain garbage bytes at the end.

Sadly, there is no way we can identify with 100% probability the end
offset of a compress'ed stream with byte precision if it is followed
by other content.

The good news is that because of this behavior, it's highly unlikely
we will observe compress'ed chunks followed by other chunks in the
wild.  The only ones I observed were followed by null bytes sentinels,
which helps identifying the exact end offset.
"""

import io

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, InvalidInputFormat, convert_int8, convert_int16
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


class UnixCompressHandler(StructHandler):
    NAME = "compress"

    PATTERNS = [
        # reference: https://fuchsia.googlesource.com/third_party/wuffs/+/HEAD/std/lzw/README.md
        HexString("1f 9d")
    ]

    C_DEFINITIONS = r"""
        struct compress_header {
            char magic[2];              // compress signature/magic number
            uint8 flags;                // blocks = flags&0x80, bits = flags&0x1f
        };
    """
    HEADER_STRUCT = "compress_header"

    EXTRACTOR = Command("7z", "x", "-y", "{inpath}", "-so", stdout="lzw.uncompressed")

    DOC = HandlerDoc(
        name=NAME,
        description="Unix compress files use the Lempel-Ziv-Welch (LZW) algorithm for data compression and are identified by a 2-byte magic number (0x1F 0x9D). This format supports optional block compression and variable bit lengths ranging from 9 to 16 bits.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="Unix Compress File Format Documentation",
                url="https://fuchsia.googlesource.com/third_party/wuffs/+/HEAD/std/lzw/README.md",
            ),
            Reference(
                title="LZW Compression Algorithm",
                url="https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch",
            ),
        ],
        limitations=[],
    )

    def unlzw(self, file: File, start_offset: int, max_len: int) -> int:  # noqa: C901
        """Calculate the end of a unix compress stream.

        It performs decompression on a stream read from <file> from
        <start_offset> up until <max_len>.

        Adapted from Brandon Owen works
        (https://github.com/umeat/unlzw).

        Adapted from original work by Mark Adler - orginal copyright
        notice below

        Copyright (C) 2014, 2015 Mark Adler This software is provided
        'as-is', without any express or implied warranty.  In no event
        will the authors be held liable for any damages arising from
        the use of this software. Permission is granted to anyone to
        use this software for any purpose, including commercial
        applications, and to alter it and redistribute it freely,
        subject to the following restrictions:

            1. The origin of this software must not be misrepresented;
               you must not claim that you wrote the original
               software.  If you use this software in a product, an
               acknowledgment in the product documentation would be
               appreciated but is not required.

            2. Altered source versions must be plainly marked as such,
               and must not be misrepresented as being the original
               software.

            3. This notice may not be removed or altered from any
               source distribution.  Mark Adler
               madler@alumni.caltech.edu
        """
        file.seek(start_offset)

        prefix: list[int] = [0] * 65536  # index to LZW prefix string

        header = self.parse_header(file, Endian.LITTLE)

        if header.flags & 0x60:
            raise InvalidInputFormat("Flag & 0x60")

        max_ = header.flags & 0x1F
        if not (9 <= max_ <= 16):
            raise InvalidInputFormat("Invalid max")

        if max_ == 9:
            max_ = 10  # 9 doesn't really mean 9

        block_compressed = header.flags & 0x80
        end = 256 if block_compressed else 255

        # Clear table, start at nine bits per symbol
        bits_per_symbol = 9
        mask = 0x1FF
        code = 0

        # Set up: get the first 9-bit code, which is the first decompressed byte,
        # but don't create a table entry until the next code

        buf = convert_int16(file.read(2), Endian.LITTLE)
        prev = buf & mask  # code
        buf >>= bits_per_symbol
        left = 16 - bits_per_symbol
        if prev > 255:
            raise InvalidInputFormat("Invalid Data: First code must be a literal")

        # Decode codes
        mark = 3  # start of compressed data
        nxt = 5  # consumed five bytes so far
        while nxt < max_len:
            # If the table will be full after this, increment the code size
            if (end >= mask) and (bits_per_symbol < max_):
                # Flush unused input bits and bytes to next 8*bits bit boundary
                # (this is a vestigial aspect of the compressed data format
                # derived from an implementation that made use of a special VAX
                # machine instruction!)
                remaining_bits = (nxt - mark) % bits_per_symbol

                if remaining_bits:
                    remaining_bits = bits_per_symbol - remaining_bits
                    if remaining_bits >= max_len - nxt:
                        break
                    nxt += remaining_bits

                buf = left = 0

                # mark this new location for computing the next flush
                mark = nxt

                # increment the number of bits per symbol
                bits_per_symbol += 1
                mask <<= 1
                mask += 1

            # Get a code of bits bits
            buf += convert_int8(file.read(1), Endian.LITTLE) << left
            nxt += 1
            left += 8
            if left < bits_per_symbol:
                if nxt == max_len:
                    raise InvalidInputFormat(
                        "Invalid Data: Stream ended in the middle of a code",
                    )
                buf += convert_int8(file.read(1), Endian.LITTLE) << left
                nxt += 1

                left += 8
            code = buf & mask
            buf >>= bits_per_symbol
            left -= bits_per_symbol

            # process clear code (256)
            if (code == 256) and block_compressed:
                # Flush unused input bits and bytes to next 8*bits bit boundary
                remaining_bits = (nxt - mark) % bits_per_symbol
                if remaining_bits:
                    remaining_bits = bits_per_symbol - remaining_bits
                    if remaining_bits > max_len - nxt:
                        break
                    nxt += remaining_bits
                buf = left = 0

                # Mark this location for computing the next flush
                mark = nxt

                # Go back to nine bits per symbol
                bits_per_symbol = 9  # initialize bits and mask
                mask = 0x1FF
                end = 255  # empty table
                continue  # get next code

            # Process LZW code
            temp = code  # save the current code

            # Special code to reuse last match
            if code > end:
                # Be picky on the allowed code here, and make sure that the
                # code we drop through (prev) will be a valid index so that
                # random input does not cause an exception
                if (code != end + 1) or (prev > end):
                    raise InvalidInputFormat("Invalid Data: Invalid code detected")
                code = prev

            # Walk through linked list to generate output in reverse order
            while code >= 256:
                code = prefix[code]

            # Link new table entry
            if end < mask:
                end += 1
                prefix[end] = prev

            # Set previous code for next iteration
            prev = temp

        if code == nxt - 1:
            return file.tell()

        return file.tell() - 1

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(0, io.SEEK_END)
        max_len = file.tell()

        end_offset = self.unlzw(file, start_offset, max_len)

        chunk_length = end_offset - start_offset
        if chunk_length <= 5:
            raise InvalidInputFormat("Compressed chunk is too short")

        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



import logging
import pdb  # noqa: T100
import sys
from logging.handlers import WatchedFileHandler
from os import getpid
from pathlib import Path
from typing import Any

import structlog
from dissect.cstruct import Structure, dumpstruct


def format_hex(value: int):
    return f"0x{value:x}"


class RawString:
    """Wrap a string so ConsoleRenderer prints it without repr()-escaping."""

    def __init__(self, value: str | None):
        self._value = value or ""

    def __repr__(self) -> str:
        return self._value

    def __str__(self) -> str:
        return self._value


class noformat:  # noqa: N801
    """Keep the value from formatting.

    Even if it would match one of the types in pretty_print_types processor.
    """

    def __init__(self, value):
        self._value = value

    def get(self):
        return self._value

    def __repr__(self) -> str:
        return repr(self._value)


def _format_message(value: Any, extract_root: Path) -> Any:
    if isinstance(value, noformat):
        return value.get()

    if isinstance(value, Path):
        try:
            new_value = value.relative_to(extract_root)
        except ValueError:
            # original files given to unblob may not be relative to extract_root
            new_value = value
        return new_value.as_posix().encode("utf-8", errors="surrogateescape")

    if isinstance(value, Structure):
        return RawString(dumpstruct(value, output="string"))

    if isinstance(value, int):
        return format_hex(value)

    if isinstance(value, str):
        try:
            value.encode()
        except UnicodeEncodeError:
            return value.encode("utf-8", errors="surrogateescape")

    return value


def pretty_print_types(extract_root: Path):
    def convert_type(_logger, _method_name: str, event_dict: structlog.types.EventDict):
        for key, value in event_dict.items():
            event_dict[key] = _format_message(value, extract_root)

        return event_dict

    return convert_type


def add_pid_to_log_message(
    _logger, _method_name: str, event_dict: structlog.types.EventDict
):
    event_dict["pid"] = getpid()
    return event_dict


def filter_debug_logs(verbosity_level: int):
    def filter_(_logger, _method_name: str, event_dict: structlog.types.EventDict):
        if event_dict["level"] != "debug":
            return event_dict

        message_verbosity: int = event_dict.pop("_verbosity", 1)
        if verbosity_level >= message_verbosity:
            return event_dict

        raise structlog.DropEvent

    return filter_


def configure_logger(verbosity_level: int, extract_root: Path, log_path: Path):
    log_path.unlink(missing_ok=True)

    log_level = logging.DEBUG if verbosity_level > 0 else logging.CRITICAL

    processors = [
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    shared_processors = [
        structlog.stdlib.add_log_level,
        filter_debug_logs(verbosity_level or 2),
        structlog.processors.TimeStamper(
            key="timestamp", fmt="%Y-%m-%d %H:%M.%S", utc=True
        ),
        pretty_print_types(extract_root),
        add_pid_to_log_message,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.StackInfoRenderer(),
    ]

    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(logging.DEBUG),
        processors=shared_processors + processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=shared_processors,
        processors=[
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            structlog.dev.ConsoleRenderer(
                colors=sys.stdout.isatty(),
                exception_formatter=structlog.dev.plain_traceback,
            ),
        ],
    )

    file_formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=shared_processors,
        processors=[
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            structlog.dev.ConsoleRenderer(
                colors=False, exception_formatter=structlog.dev.plain_traceback
            ),
        ],
    )

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    console_handler.setLevel(log_level)

    file_handler = WatchedFileHandler(log_path.as_posix())
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)

    root_logger = logging.getLogger()
    root_logger.addHandler(console_handler)
    root_logger.addHandler(file_handler)
    root_logger.setLevel(logging.DEBUG)
    structlog.get_logger().debug(
        "Logging configured",
        vebosity_level=noformat(verbosity_level),
        extract_root=extract_root.expanduser().resolve(),
    )


class _MultiprocessingPdb(pdb.Pdb):
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        with Path("/dev/stdin").open() as new_stdin:
            try:
                sys.stdin = new_stdin
                pdb.Pdb.interaction(self, *args, **kwargs)
            finally:
                sys.stdin = _stdin


def multiprocessing_breakpoint():
    """Call this in Process forks instead of the builtin `breakpoint` function for debugging with PDB."""
    return _MultiprocessingPdb().set_trace(frame=sys._getframe(1))  # noqa: SLF001



import re
import zlib
from pathlib import Path

from structlog import get_logger

from unblob.handlers.archive.dmg import DMGHandler

from ...file_utils import DEFAULT_BUFSIZE, InvalidInputFormat
from ...models import (
    Extractor,
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    ValidChunk,
)

logger = get_logger()


class ZlibExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        decompressor = zlib.decompressobj()
        outpath = outdir / "zlib.uncompressed"
        with File.from_path(inpath) as f, outpath.open("wb") as outfile:
            content = f.read(DEFAULT_BUFSIZE)
            while content and not decompressor.eof:
                outfile.write(decompressor.decompress(content))
                content = f.read(DEFAULT_BUFSIZE)


class ZlibHandler(Handler):
    NAME = "zlib"

    PATTERNS = [
        Regex(r"^\x78\x01"),  # low compression
        Regex(r"^\x78\x9c"),  # default compression
        Regex(r"^\x78\xda"),  # best compression
        Regex(r"^\x78\x5e"),  # compressed
    ]

    EXTRACTOR = ZlibExtractor()

    DOC = HandlerDoc(
        name=NAME,
        description="The zlib format is a compressed data format based on the DEFLATE algorithm, often used for data compression in various applications. It includes a lightweight header and checksum for data integrity.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="zlib File Format Specification",
                url="https://www.zlib.net/manual.html",
            ),
            Reference(
                title="zlib Wikipedia",
                url="https://en.wikipedia.org/wiki/Zlib",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        for pattern in DMGHandler.PATTERNS:
            if re.search(pattern.as_regex(), file[-512:]):
                raise InvalidInputFormat(
                    "File is a DMG archive made of zlib streams. Aborting."
                )

        decompressor = zlib.decompressobj()

        try:
            content = file.read(DEFAULT_BUFSIZE)
            while content and not decompressor.eof:
                decompressor.decompress(content)
                content = file.read(DEFAULT_BUFSIZE)

        except zlib.error:
            raise InvalidInputFormat("invalid zlib stream") from None

        end_offset = file.tell() - len(decompressor.unused_data)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



import io

from ...extractors import Command
from ...file_utils import Endian
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

PADDING_LEN = 2
# CPP/7zip/Archive/LzhHandler.cpp
HEADER_MIN_SIZE = 2 + 22


class LZHHandler(StructHandler):
    NAME = "lzh"

    PATTERNS = [
        Regex(r"-lh0-"),
        Regex(r"-lzs-"),
        Regex(r"-lz4-"),
        Regex(r"-lh1-"),
        Regex(r"-lh2-"),
        Regex(r"-lh3-"),
        Regex(r"-lh4-"),
        Regex(r"-lh5-"),
        Regex(r"-lh6-"),
        Regex(r"-lh7-"),
        Regex(r"-lh8-"),
        Regex(r"-lhd-"),
    ]

    PATTERN_MATCH_OFFSET = -2

    C_DEFINITIONS = r"""
        typedef struct lzh_default_header {
            uint8 header_size;          // excludes extended headers size
            uint8 header_checksum;
            char method_id[5];
            uint32 compressed_size;     // includes all extended headers size (if level 1)
            uint32 uncompressed_size;
            uint32 timestamp;
            uint8 fd_attribute;
            uint8 level_identifier;
        } lzh_default_header_t;

        typedef struct level_2_header {
            uint16 header_size; // includes all extended headers
            char method_id[5];
            uint32 compressed_size;     // excludes all extended headers
            uint32 uncompressed_size;
            uint32 timestamp;
            uint8 fd_attribute;
            uint8 level_identifier;
        } level_2_header_t;
    """
    HEADER_STRUCT = "lzh_default_header_t"

    EXTRACTOR = Command("7z", "x", "-p", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="LZH",
        description="LZH is a legacy archive format that uses various compression methods such as '-lh0-' and '-lh5-'. It was widely used in Japan and on older systems for compressing and archiving files.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="LZH Compression Format",
                url="https://en.wikipedia.org/wiki/LHA_(file_format)",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, Endian.LITTLE)

        if header.level_identifier > 0x2:
            return None

        if header.level_identifier == 0x2:
            # with level 2, the header size is a uint16 rather than uint8 and there
            # is no checksum. We use this magic trick so we don't parse the header
            # again. See the level_2_header definition in C_DEFINITIONS
            header_size = header.header_size + (header.header_checksum << 8)
        else:
            header_size = header.header_size + PADDING_LEN

        if header_size < HEADER_MIN_SIZE:
            return None

        file.seek(-len(header), io.SEEK_CUR)
        file.seek(header_size + header.compressed_size, io.SEEK_CUR)
        end_offset = file.tell()

        # LZH files are null terminated, so we have to handle the case where
        # we matched the last LZH stream of a file and pad appropriately.
        file.seek(0, io.SEEK_END)
        end_pos = file.tell()

        if end_pos - end_offset == 1:
            end_offset = end_pos

        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



import importlib.util
import itertools
import sys
from pathlib import Path
from types import ModuleType

import pluggy
from structlog import get_logger

from unblob import hookspecs
from unblob.models import DirectoryHandler, Handler
from unblob.report import Report

# The entrypoints are defined by the to-be-loaded plugins. The version
# should be incremented whenever a backward-incompatible change is
# introduced.
SETUPTOOLS_ENTRYPOINT_NAME = "unblob_v1"

logger = get_logger()

hookimpl = pluggy.HookimplMarker("unblob")


class UnblobPluginManager(pluggy.PluginManager):
    def __init__(self):
        super().__init__("unblob")
        self.add_hookspecs(hookspecs)

    def load_setuptools_entrypoints(
        self, group: str = SETUPTOOLS_ENTRYPOINT_NAME, name: str | None = None
    ):
        return super().load_setuptools_entrypoints(group, name=name)

    def import_path(self, path: Path):
        """Load Python code from a given path.

        The following scenarios are supported based on the contents of
        ``path``:

        Single file
        -----------

        Path points to a single Python file

        Multiple files
        --------------

        A directory containing one or more Python files or packages.
        Files are loaded from the directory root.  Additional
        sub-directories containing Python packages (directories with
        ``__init__.py``) are loaded recursively.
        """
        logger.debug("Importing plugin modules", path=path)

        if path.is_file():
            module_name = path.stem
            to_import = [(module_name, path)]
        elif path.is_dir():
            to_import = [(p.parent.name, p) for p in path.glob("*/__init__.py")]
            to_import.extend((p.stem, p) for p in path.glob("*.py"))
        else:
            raise ValueError("Invalid plugin import path", path)

        modules = self._import_modules(to_import)
        logger.debug("Imported plugins", modules=modules)
        for module in modules:
            self.register(module)

    @classmethod
    def _import_modules(
        cls, modules_to_import: list[tuple[str, Path]]
    ) -> list[ModuleType]:
        modules = []
        for module_name, path in modules_to_import:
            spec = importlib.util.spec_from_file_location(module_name, path)
            if not spec or not spec.loader:
                logger.error("Invalid plugin file", path=path)
                continue

            mod = importlib.util.module_from_spec(spec)
            # This is crucial for package loading, in order to module
            # import inside a package to work.
            sys.modules[module_name] = mod
            spec.loader.exec_module(mod)
            modules.append(mod)

        return modules

    def import_plugins(self, path: Path | None = None):
        if path:
            self.import_path(path)

        self.load_setuptools_entrypoints()
        plugins = [name for name, _plugin in self.list_name_plugin()]
        if not plugins:
            return

        logger.info("Loaded plugins", plugins=plugins)

    def load_handlers_from_plugins(
        self,
    ) -> list[type[Handler]]:
        extra_handlers = list(itertools.chain(*self.hook.unblob_register_handlers()))  # type: ignore
        if extra_handlers:
            logger.debug("Loaded handlers from plugins", handlers=extra_handlers)

        return extra_handlers

    def load_dir_handlers_from_plugins(
        self,
    ) -> list[type[DirectoryHandler]]:
        extra_handlers = list(
            itertools.chain(*self.hook.unblob_register_dir_handlers())
        )  # type: ignore
        if extra_handlers:
            logger.debug(
                "Loaded directory handlers from plugins", handlers=extra_handlers
            )

        return extra_handlers

    def load_report_types_from_plugins(self) -> list[type[Report]]:
        extra_report_types = list(itertools.chain(*self.hook.unblob_register_reports()))
        if extra_report_types:
            logger.debug(
                "Loaded report types from plugins", report_types=extra_report_types
            )
        return extra_report_types



import hashlib
import io
from pathlib import Path

from unblob.file_utils import Endian, StructParser
from unblob.models import (
    DirectoryHandler,
    Glob,
    HandlerDoc,
    HandlerType,
    MultiFile,
    Reference,
)

C_DEFINITIONS = r"""
    typedef struct par2_header{
        char magic[8];
        uint64 packet_length;
        char md5_hash[16];
        char recovery_set_id[16];
        char type[16];
    } par2_header_t;
"""

PAR2_MAGIC = b"PAR2\x00PKT"
HEADER_STRUCT = "par2_header_t"
HEADER_PARSER = StructParser(C_DEFINITIONS)


class MultiVolumePAR2Handler(DirectoryHandler):
    NAME = "multi-par2"
    PATTERN = Glob("*.par2")
    EXTRACTOR = None

    DOC = HandlerDoc(
        name="PAR2 (multi-volume)",
        description="Parchive or PAR2, is a format for creating redundant data that helps detect and repair corrupted files. These archives typically accompany split-file sets (like multi-volume RAR or ZIP archives). Each PAR2 file is composed of multiple 'packets'.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="Parchive Documentation",
                url="https://parchive.github.io/",
            ),
        ],
        limitations=[],
    )

    def is_valid_header(self, file_paths: list) -> bool:
        for path in file_paths:
            with path.open("rb") as f:
                header = HEADER_PARSER.parse(HEADER_STRUCT, f, Endian.LITTLE)
                if header.magic != PAR2_MAGIC:
                    return False

                offset_to_recovery_id = 32
                # seek to beginning of recovery set ID
                f.seek(offset_to_recovery_id, io.SEEK_SET)
                packet_content = f.read(
                    header.packet_length - len(header) + offset_to_recovery_id
                )
                packet_checksum = hashlib.md5(
                    packet_content, usedforsecurity=False
                ).digest()

                if packet_checksum != header.md5_hash:
                    return False
        return True

    def calculate_multifile(self, file: Path) -> MultiFile | None:
        paths = sorted(
            [p for p in file.parent.glob(f"{file.stem}.*") if p.resolve().exists()]
        )

        if len(paths) <= 1 or not self.is_valid_header(paths):
            return None

        return MultiFile(
            name=file.stem,
            paths=paths,
        )



from __future__ import annotations

import io
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from struct import Struct
from typing import TYPE_CHECKING

from unblob.file_utils import Endian, FileSystem, InvalidInputFormat, StructParser
from unblob.models import (
    Extractor,
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    ValidChunk,
)

if TYPE_CHECKING:
    from collections.abc import Iterable


C_DEFINITIONS = r"""
    typedef struct cpkg_header {
        char     magic[4];           // "CPKG"
        uint32   unknown;            // 0x01_00_00_00 in all samples (maybe version)
        uint32   first_entry_offset;
    } cpkg_header_t;

    typedef struct fpkg_header {
        char     magic[4];           // "FPKG"
        uint32   unknown;            // 0x01_00_00_00 in all samples
        uint32   first_entry_offset;
        uint32   unknown2;           // only 0x00_00_00_01 observed
        uint32   name_len;           // length of model name field
        char     name[name_len];
    } fpkg_header_t;

    typedef struct file_header {
        uint32   header_len;         // only 0x1C (28) observed
        uint16   type;               // probably file type
        uint16   unknown;            // observed values: 0x0 and 0x6874
        uint32   file_size;          // size of file data
        char     filename[];
    } file_header_t;
"""
CPKG_HEADER = "cpkg_header_t"
FILE_HEADER = "file_header_t"
CPKG_HEADER_SIZE = 12
FILE_HEADER_SIZE = 0x1C
VALID_MAGICS = {b"FPKG", b"CPKG"}


@dataclass
class FileHeader(Struct):
    header_len: int
    type: int
    start_offset: int
    file_size: int
    filename: str
    total_size: int


class FileType(IntEnum):
    REGULAR_FILE = 0x100
    UNKNOWN = 0x101  # NOTE: not observed in the wild, but makes sense in sequence
    CHECKSUM = 0x102
    SIGNATURE = 0x103


class FPKGParser:
    def __init__(self, file: File, start_offset: int = 0):
        self.file = file
        self.start_offset = start_offset
        self.struct_parser = StructParser(C_DEFINITIONS)
        self.header = self.struct_parser.parse(CPKG_HEADER, file, Endian.BIG)
        self._validate_header(self.header)
        self.file_data_offset = start_offset + self.header.first_entry_offset

    def iter_entries(self) -> Iterable[FileHeader]:
        current_offset = self.file_data_offset
        while current_offset < self.file.size():
            self.file.seek(current_offset, io.SEEK_SET)
            entry = self.struct_parser.parse(FILE_HEADER, self.file, Endian.BIG)
            self._validate_file_header(entry)
            yield FileHeader(
                header_len=entry.header_len,
                file_size=entry.file_size,
                start_offset=current_offset + entry.header_len,
                type=entry.type,
                filename=entry.filename.rstrip().decode("utf-8", errors="replace"),
                total_size=entry.file_size + entry.header_len,
            )
            current_offset += entry.header_len + entry.file_size

    @staticmethod
    def _validate_file_header(file_header) -> None:
        if file_header.header_len != FILE_HEADER_SIZE:
            raise InvalidInputFormat(
                f"Invalid file header length: {file_header.header_len}"
            )
        if not file_header.filename.isascii():
            raise InvalidInputFormat(f"Invalid filename: {file_header.filename}")
        try:
            FileType(file_header.type)
        except ValueError as e:
            raise InvalidInputFormat(f"Invalid file type: {file_header.type}") from e

    @staticmethod
    def _validate_header(header) -> None:
        if header.magic not in VALID_MAGICS:
            raise InvalidInputFormat("Invalid magic")
        if header.first_entry_offset < CPKG_HEADER_SIZE:
            raise InvalidInputFormat(
                f"Invalid first entry offset: {header.first_entry_offset}"
            )


class FPKGExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            parser = FPKGParser(file)
            for entry in parser.iter_entries():
                fs.carve(
                    Path(entry.filename), file, entry.start_offset, entry.file_size
                )


class FPKGHandler(Handler):
    NAME = "fpkg"
    PATTERNS = [HexString("(43 | 46) 50 4B 47 01 00 00 00")]  # (C | F) P K G
    # FPKG and CPKG headers are compatible if we only care about the "first_entry_offset"
    EXTRACTOR = FPKGExtractor()
    DOC = HandlerDoc(
        name="D-Link FPKG",
        description="CPKG and FPKG are archive formats used in D-Link DFL firewall firmware",
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        parser = FPKGParser(file, start_offset)
        entries = list(parser.iter_entries())
        if not entries:
            raise InvalidInputFormat("No valid entries found")

        end_offset = parser.file_data_offset + sum(e.total_size for e in entries)
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



"""File extraction related functions."""

import errno
import os
from pathlib import Path

from structlog import get_logger

from .file_utils import carve, is_safe_path
from .models import Chunk, File, PaddingChunk, TaskResult, UnknownChunk, ValidChunk
from .report import MaliciousSymlinkRemoved

logger = get_logger()

FILE_PERMISSION_MASK = 0o644
DIR_PERMISSION_MASK = 0o775


def carve_chunk_to_file(carve_path: Path, file: File, chunk: Chunk):
    """Extract valid chunk to a file, which we then pass to another tool to extract it."""
    logger.debug("Carving chunk", path=carve_path)
    carve(carve_path, file, chunk.start_offset, chunk.size)


def fix_permission(path: Path):
    if path.is_symlink():
        return

    if not path.exists():
        return

    mode = path.stat().st_mode

    if path.is_file():
        mode |= FILE_PERMISSION_MASK
    elif path.is_dir():
        mode |= DIR_PERMISSION_MASK

    path.chmod(mode)


def is_recursive_link(path: Path) -> bool:
    try:
        path.resolve()
    except RuntimeError:
        return True
    return False


def fix_symlink(path: Path, outdir: Path, task_result: TaskResult) -> Path:
    """Rewrites absolute symlinks to point within the extraction directory (outdir).

    If it's not a relative symlink it is either removed it it attempts
    to traverse outside of the extraction directory or rewritten to be
    fully portable (no mention of the extraction directory in the link
    value).
    """
    if is_recursive_link(path):
        logger.error("Symlink loop identified, removing", path=path)
        error_report = MaliciousSymlinkRemoved(
            link=path.as_posix(), target=path.readlink().as_posix()
        )
        task_result.add_report(error_report)
        path.unlink()
        return path

    raw_target = os.readlink(path)  # noqa: PTH115
    if not raw_target:
        logger.error("Symlink with empty target, removing.")
        path.unlink()
        return path

    target = Path(raw_target)
    if target.is_absolute():
        target = Path(target.as_posix().lstrip("/"))
    else:
        target = path.resolve()

    safe = is_safe_path(outdir, target)

    if not safe:
        logger.error("Path traversal attempt through symlink, removing", target=target)
        error_report = MaliciousSymlinkRemoved(
            link=path.as_posix(), target=target.as_posix()
        )
        task_result.add_report(error_report)
        path.unlink()
    else:
        relative_target = os.path.relpath(outdir.joinpath(target), start=path.parent)
        path.unlink()
        path.symlink_to(relative_target)
    return path


def fix_extracted_directory(outdir: Path, task_result: TaskResult):
    def _fix_extracted_directory(directory: Path):
        if not directory.exists():
            return
        for path in directory.iterdir():
            try:
                fix_permission(path)
                if path.is_symlink():
                    fix_symlink(path, outdir, task_result)
                    continue
                if path.is_dir():
                    _fix_extracted_directory(path)
            except OSError as e:
                if e.errno == errno.ENAMETOOLONG:
                    continue
                raise e from None

    fix_permission(outdir)
    _fix_extracted_directory(outdir)


def carve_unknown_chunk(
    extract_dir: Path, file: File, chunk: UnknownChunk | PaddingChunk
) -> Path:
    extension = "unknown"
    if isinstance(chunk, PaddingChunk):
        extension = "padding"

    filename = f"{chunk.start_offset}-{chunk.end_offset}.{extension}"
    carve_path = extract_dir / filename
    logger.info("Extracting unknown chunk", path=carve_path, chunk=chunk)
    carve_chunk_to_file(carve_path, file, chunk)
    return carve_path


def carve_valid_chunk(extract_dir: Path, file: File, chunk: ValidChunk) -> Path:
    filename = f"{chunk.start_offset}-{chunk.end_offset}.{chunk.handler.NAME}"
    carve_path = extract_dir / filename
    logger.info("Extracting valid chunk", path=carve_path, chunk=chunk)
    carve_chunk_to_file(carve_path, file, chunk)
    return carve_path



import binascii
import io
from math import ceil

from unblob.extractors import Command
from unblob.file_utils import File, InvalidInputFormat, get_endian_short
from unblob.models import (
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

C_DEFINITIONS = r"""
    typedef struct partclone_header{
        char magic[16];
        char partclone_version[14];
        char image_version_txt[4];
        char endian[2];
        char fs_type[16];
        uint64 fs_size;
        uint64 fs_total_block_count;
        uint64 fs_used_block_count_superblock;
        uint64 fs_used_block_count_bitmap;
        uint32 fs_block_size;
        uint32 feature_size;
        uint16 image_version;
        uint16 number_of_bits_for_CPU;
        uint16 checksum_mode;
        uint16 checksum_size;
        uint32 blocks_per_checksum;
        uint8 reseed_checksum;
        uint8 bitmap_mode;
        uint32 crc32;
    } partclone_header_t;
"""

HEADER_STRUCT = "partclone_header_t"
BIG_ENDIAN_MAGIC = 0xC0DE
ENDIAN_OFFSET = 34


class PartcloneHandler(StructHandler):
    NAME = "partclone"
    PATTERNS = [Regex(r"partclone-image\x00\d+\.\d+\.\d+.*?0002(\xde\xc0|\xc0\xde)")]
    HEADER_STRUCT = HEADER_STRUCT
    C_DEFINITIONS = C_DEFINITIONS
    EXTRACTOR = Command(
        "partclone.restore",
        "-W",
        "-s",
        "{inpath}",
        "-o",
        "{outdir}/partclone.restored",
        "-L",
        "/dev/stdout",
    )
    DOC = HandlerDoc(
        name="Partclone",
        description="Partclone is a utility used for backing up and restoring partitions. Many cloning tools (such as Clonezilla) rely on it to create block-level images that include filesystem metadata.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="Partclone GitHub Repository",
                url="https://github.com/Thomas-Tsai/partclone",
            ),
            Reference(
                title="Clonezilla Official Documentation",
                url="https://clonezilla.org/",
            ),
        ],
        limitations=[],
    )

    def is_valid_header(self, header) -> bool:
        calculated_crc = binascii.crc32(header.dumps()[0:-4])
        return (
            header.crc32 ^ 0xFFFFFFFF
        ) == calculated_crc  # partclone does not final XOR

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset + ENDIAN_OFFSET, io.SEEK_SET)  # go to endian
        endian = get_endian_short(file, BIG_ENDIAN_MAGIC)
        file.seek(start_offset, io.SEEK_SET)  # go to beginning of file
        header = self.parse_header(file, endian)

        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid partclone header.")

        end_offset = start_offset + len(header)  # header
        end_offset += header.checksum_size  # checksum size
        end_offset += ceil(header.fs_total_block_count / 8)  # bitmap, as bytes

        if header.checksum_mode != 0:
            checksum_blocks = ceil(
                header.fs_used_block_count_bitmap / header.blocks_per_checksum
            )
            end_offset += checksum_blocks * header.checksum_size

        end_offset += header.fs_used_block_count_bitmap * header.fs_block_size  # Data
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



from __future__ import annotations

import gzip
import io
from pathlib import Path, PureWindowsPath

from unblob.file_utils import Endian, File, FileSystem, InvalidInputFormat, StructParser
from unblob.models import (
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    StructHandler,
    ValidChunk,
)

C_DEFINITIONS = r"""
    typedef struct deafbead_header {
        uint32 magic;           /* "DE AF BE AD" */
    } deafbead_header_t;

    typedef struct deafbead_dir {
        uint8  magic;           /* "86" */
        uint16 name_len;
        char   name[name_len];
    } deafbead_dir_t;

    typedef struct deafbead_file {
        uint8  magic;           /* "87" */
        uint16 name_len;
        char   name[name_len];
        uint32 file_size;
        char   file_contents[file_size];
    } deafbead_file_t;
"""
DIR_MAGIC = b"\x86"
FILE_MAGIC = b"\x87"
VALID_MAGICS = {DIR_MAGIC, FILE_MAGIC}
HEADER_LEN = 4


class DeafBeadExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            file.seek(HEADER_LEN)
            while (magic := file.read(1)) in VALID_MAGICS:
                file.seek(-1, io.SEEK_CUR)  # go back to read the full struct
                if magic == DIR_MAGIC:
                    self._handle_dir(file, fs)
                elif magic == FILE_MAGIC:
                    self._handle_file(file, fs)

    def _handle_dir(self, file: File, fs: FileSystem):
        dir_header = self._struct_parser.parse("deafbead_dir_t", file, Endian.LITTLE)
        fs.mkdir(self._convert_path(dir_header.name))

    def _handle_file(self, file: File, fs: FileSystem):
        file_header = self._struct_parser.parse("deafbead_file_t", file, Endian.LITTLE)
        try:
            decompressed = gzip.decompress(file_header.file_contents)
            fs.write_bytes(self._convert_path(file_header.name), decompressed)
        except gzip.BadGzipFile as error:
            raise InvalidInputFormat("Invalid GZIP file") from error

    @staticmethod
    def _convert_path(path_entry: bytes) -> Path:
        decoded_path = path_entry.decode("utf-8", errors="replace")
        if "\\" in decoded_path:  # windows path => convert slashes
            return Path(PureWindowsPath(decoded_path).as_posix())
        return Path(decoded_path)


class DeafBeadHandler(StructHandler):
    NAME = "deafbead"
    PATTERNS = [HexString("DE AF BE AD (86 | 87)")]

    DOC = HandlerDoc(
        name="D-Link DEAFBEAD",
        description="Archive files as found in D-Link DSL-500G and DSL-504G firmware images.",
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[],
        limitations=[],
    )

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "deafbead_header_t"
    EXTRACTOR = DeafBeadExtractor()

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset + HEADER_LEN)
        while (magic := file.read(1)) in VALID_MAGICS:
            file.seek(-1, io.SEEK_CUR)
            if magic == DIR_MAGIC:
                header = self.cparser_le.deafbead_dir_t(file)
                if header.name_len == 0:
                    raise InvalidInputFormat("Invalid directory header.")
            else:
                header = self.cparser_le.deafbead_file_t(file)
                if header.name_len == 0 or header.file_size == 0:
                    raise InvalidInputFormat("Invalid file header.")

        end_offset = file.tell()
        if magic:  # if EOF wasn't reached (i.e. magic is not empty), we need to undo the last read
            end_offset -= 1
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



import binascii
import io
from collections.abc import Iterable
from pathlib import Path
from typing import cast

from structlog import get_logger

from unblob.file_utils import (
    File,
    FileSystem,
    InvalidInputFormat,
    iterate_file,
    snull,
)
from unblob.models import (
    Endian,
    Extractor,
    ExtractResult,
    HandlerDoc,
    HandlerType,
    HexString,
    StructHandler,
    StructParser,
    ValidChunk,
)

logger = get_logger()
CRC_CONTENT_OFFSET = 12  # The CRC32 value is located after 12 byte in the header
SIGNATURE_LEN = 272  # signature_header_t contains 4 bytes of size + 12 bytes for padding x3 + 0x100 is 256 in decimal
BLOB_MAGIC = 0x000000BEBA  # Blob header magic

# https://lxr.openwrt.org/source/firmware-utils/src/xiaomifw.c
C_DEFINITIONS = r"""
    struct hdr1_header {
          char magic[4];                /* HDR1 */
          uint32 signature_offset;
          uint32 crc32;
          uint16 unused;
          uint16 device_id;             /* RA70 */
          uint32 blob_offsets[8];
    } hdr1_header_t;

    struct hdr2_header {
          char magic[4];                /* HDR1 */
          uint32 signature_offset;
          uint32 crc32;
          uint32 unused1;
          uint64 device_id;             /* RA70 */
          uint64 region;                /* EU */
          uint64 unused2[2];
          uint32 blob_offsets[8];
    } hdr2_header_t;

    struct xiaomi_blob_header {
          uint32 magic;                 /* 0x0000babe */
          uint32 flash_offset;
          uint32 blob_size;             /* Size of blob */
          uint16 type;                  /* Type of blob */
          uint16 unused;
          char name[32];                /* Name of blob */
    } blob_header_t;

    struct xiaomi_signature_header {
          uint32 signature_size;
          uint32 padding[3];
          uint8 content[0x100];
    } signature_header_t;
 """


def calculate_crc(file: File, start_offset: int, size: int) -> int:
    digest = 0
    for chunk in iterate_file(file, start_offset, size):
        digest = binascii.crc32(chunk, digest)
    return (digest ^ -1) & 0xFFFFFFFF


def is_valid_blob_header(blob_header) -> bool:
    if blob_header.magic == BLOB_MAGIC:
        return False
    if not blob_header.blob_size:
        return False
    try:
        snull(blob_header.name).decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


def is_valid_header(header) -> bool:
    if header.signature_offset < len(header):
        return False
    if not header.blob_offsets[0]:  # noqa: SIM103
        return False
    return True


class HDRExtractor(Extractor):
    def __init__(self, header_struct: str):
        self.header_struct = header_struct
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            for output_path, start_offset, size in self.parse(file):
                fs.carve(output_path, file, start_offset, size)
        return ExtractResult(reports=fs.problems)

    def parse(self, file: File) -> Iterable[tuple[Path, int, int]]:
        header = self._struct_parser.parse(self.header_struct, file, Endian.LITTLE)
        for offset in cast("Iterable", header.blob_offsets):
            if not offset:
                break

            file.seek(offset, io.SEEK_SET)
            blob_header = self._struct_parser.parse(
                "blob_header_t", file, Endian.LITTLE
            )
            logger.debug("blob_header_t", blob_header_t=blob_header, _verbosity=3)
            if not is_valid_blob_header(blob_header):
                raise InvalidInputFormat("Invalid HDR blob header.")

            yield (
                (
                    Path(snull(blob_header.name).decode("utf-8")),
                    # file.tell() points to right after the blob_header == start_offset
                    file.tell(),
                    blob_header.blob_size,
                )
            )


class HDRHandlerBase(StructHandler):
    HEADER_STRUCT = "hdr1_header_t"

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, endian=Endian.LITTLE)

        if not is_valid_header(header):
            raise InvalidInputFormat("Invalid HDR header.")

        end_offset = start_offset + header.signature_offset + SIGNATURE_LEN

        if not self._is_crc_valid(file, header, start_offset, end_offset):
            raise InvalidInputFormat("CRC32 does not match in HDR header.")

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)

    def _is_crc_valid(
        self, file: File, header, start_offset: int, end_offset: int
    ) -> bool:
        computed_crc = calculate_crc(
            file,
            start_offset=start_offset + CRC_CONTENT_OFFSET,
            size=end_offset - start_offset + CRC_CONTENT_OFFSET,
        )
        return header.crc32 == computed_crc


class HDR1Handler(HDRHandlerBase):
    NAME = "hdr1"
    PATTERNS = [
        HexString(
            """
            //  48 44 52 31 90 32 e2 00  02 2a 5b 6a 00 00 11 00  HDR1.2...*[j....
            //  30 00 00 00 70 02 00 00  a4 02 e0 00 00 00 00 00  0...p...........
            48 44 52 31
        """
        ),
    ]
    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "hdr1_header_t"
    EXTRACTOR = HDRExtractor("hdr1_header_t")

    DOC = HandlerDoc(
        name="Xiaomi HDR1",
        description="Xiaomi HDR1 firmware files feature a custom header containing metadata, CRC32 checksum, and blob offsets for embedded data. These files are used in Xiaomi devices for firmware updates.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Xiaomi",
        references=[],
        limitations=[],
    )


class HDR2Handler(HDRHandlerBase):
    NAME = "hdr2"
    PATTERNS = [
        HexString(
            """
            // 48 44 52 32 d4 02 78 02  68 54 e8 fa 00 00 00 00  HDR2..x.hT......
            // 52 41 37 30 00 00 00 00  45 55 00 00 00 00 00 00  RA70....EU......
            48 44 52 32
        """
        ),
    ]
    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "hdr2_header_t"
    EXTRACTOR = HDRExtractor("hdr2_header_t")

    DOC = HandlerDoc(
        name="Xiaomi HDR2",
        description="Xiaomi HDR2 firmware files feature a custom header with metadata, CRC32 checksum, and blob offsets for embedded data. These files also include additional fields for device ID and region information.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Xiaomi",
        references=[],
        limitations=[],
    )



import io
import zlib
from enum import IntEnum

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, convert_int32
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

MAGIC_LENGTH = 9
CHECKSUM_LENGTH = 4


# Header flags defined in lzop (http://www.lzop.org/) source in src/conf.h
class HeaderFlags(IntEnum):
    ADLER32_D = 0x00000001
    ADLER32_C = 0x00000002
    STDIN = 0x00000004
    STDOUT = 0x00000008
    NAME_DEFAULT = 0x00000010
    DOSISH = 0x00000020
    H_EXTRA_FIELD = 0x00000040
    H_GMTDIFF = 0x00000080
    CRC32_D = 0x00000100
    CRC32_C = 0x00000200
    MULTIPART = 0x00000400
    H_FILTER = 0x00000800
    H_CRC32 = 0x00001000
    H_PATH = 0x00002000


class LZOHandler(StructHandler):
    NAME = "lzo"

    PATTERNS = [HexString("89 4C 5A 4F 00 0D 0A 1A 0A")]

    C_DEFINITIONS = r"""
        typedef struct lzo_header_no_filter
        {
            char magic[9];
            uint16 version;
            uint16 libversion;
            uint16 reqversion;
            uint8 method;
            uint8 level;
            uint32 flags;
            //uint32 filter;              // only if flags & F_H_FILTER
            uint32 mode;
            uint32 mtime;
            uint32 gmtdiff;
            uint8 filename_len;
            char filename[filename_len];
            uint32 header_checksum;        // (CRC32 if flags & F_H_CRC32 else Adler32)
        } lzo_header_no_filter_t;

        typedef struct lzo_header_filter
        {
            char magic[9];
            uint16 version;
            uint16 libversion;
            uint16 reqversion;
            uint8 method;
            uint8 level;
            uint32 flags;
            uint32 filter;              // only if flags & F_H_FILTER
            uint32 mode;
            uint32 mtime;
            uint32 gmtdiff;
            uint8 filename_len;
            char filename[filename_len];
            uint32 header_checksum;        // (CRC32 if flags & F_H_CRC32 else Adler32)
        } lzo_header_filter_t;
    """
    HEADER_STRUCT = "lzo_header"

    EXTRACTOR = Command("lzop", "-d", "-f", "-f", "-N", "-p{outdir}", "{inpath}")

    DOC = HandlerDoc(
        name="LZO",
        description="LZO is a data compression format featuring a simple header structure and optional checksum verification. It is optimized for fast decompression and supports various compression levels and flags for additional metadata.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="LZO File Format Documentation",
                url="http://www.lzop.org/",
            ),
            Reference(
                title="LZO Wikipedia",
                url="https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Oberhumer",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.cparser_be.lzo_header_no_filter_t(file)
        # maxmimum compression level is 9
        if header.level > 9:
            logger.debug("Invalid LZO header level", header=header, _verbosity=3)
            return None

        if header.flags & HeaderFlags.H_FILTER:
            file.seek(start_offset)
            header = self.cparser_be.lzo_header_filter_t(file)

        logger.debug("LZO header parsed", header=header, _verbosity=3)

        # Checksum excludes the magic and the checksum itself
        if header.flags & HeaderFlags.H_CRC32:
            calculated_checksum = zlib.crc32(
                header.dumps()[MAGIC_LENGTH:-CHECKSUM_LENGTH]
            )
        else:
            calculated_checksum = zlib.adler32(
                header.dumps()[MAGIC_LENGTH:-CHECKSUM_LENGTH]
            )

        if header.header_checksum != calculated_checksum:
            logger.debug("Header checksum verification failed")
            return None

        uncompressed_size = convert_int32(file.read(4), endian=Endian.BIG)
        while uncompressed_size:
            compressed_size = convert_int32(file.read(4), endian=Endian.BIG)

            checksum_size = 0
            if (
                header.flags & HeaderFlags.ADLER32_D
                or header.flags & HeaderFlags.CRC32_D
            ):
                checksum_size += CHECKSUM_LENGTH

            if (
                header.flags & HeaderFlags.ADLER32_C
                or header.flags & HeaderFlags.CRC32_C
            ):
                checksum_size += CHECKSUM_LENGTH

            file.seek(checksum_size + compressed_size, io.SEEK_CUR)
            uncompressed_size = convert_int32(file.read(4), endian=Endian.BIG)

        end_offset = file.tell()

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



# Shared infrastructure for QNAP firmware handlers.
import io
from pathlib import Path

from unblob.file_utils import File, FileSystem, iterate_file
from unblob.models import (
    Endian,
    Extractor,
    ExtractResult,
    StructParser,
)

FOOTER_LEN = 74

NAS_DEVICE_ID_PREFIX = "QNAPNAS"

C_DEFINITIONS = """
    typedef struct qnap_header {
        char    magic[6];
        uint32  encrypted_len;
        char    device_id[16];
        char    file_version[16];
        char    firmware_date[16];
        char    revision[16];
    } qnap_header_t;
"""


# https://gist.github.com/ulidtko/966277a465f1856109b2d2674dcee741#file-qnap-qts-fw-cryptor-py-L114
class Cryptor:
    def __init__(self, secret):
        self.secret = list(bytes(secret, "ascii"))
        self.n = len(secret) // 2
        if self.n % 2 == 0:
            self.secret.append(0)
        self.precompute_k()
        self.acc = 0
        self.y = 0
        self.z = 0

    def scan(self, f, xs, s0):
        s = s0
        for x in xs:
            w, s = f(s, x)
            yield w

    def promote(self, char):
        return char if char < 0x80 else char - 0x101

    def precompute_k(self):
        self.k = {acc: self.table_for_acc(acc) for acc in range(256)}

    def table_for_acc(self, a):
        ks = [
            0xFFFFFFFF
            & (
                (self.promote(self.secret[2 * i] ^ a) << 8)
                + (self.secret[2 * i + 1] ^ a)
            )
            for i in range(self.n)
        ]

        def kstep(st, q):
            x = st ^ q
            y = self.lcg(x)
            z = 0xFFFF & (0x15A * x)
            return (z, y), y

        return list(self.scan(kstep, ks, 0))

    def lcg(self, x):
        return 0xFFFF & (0x4E35 * x + 1)

    def kdf(self):
        """self.secret -> 8bit hash (+ state effects)."""
        tt = self.k[self.acc]
        res = 0
        for i in range(self.n):
            yy = self.y
            self.y, t2 = tt[i]
            self.z = 0xFFFF & (self.y + yy + 0x4E35 * (self.z + i))
            res = res ^ t2 ^ self.z
        hi, lo = res >> 8, res & 0xFF
        return hi ^ lo

    def decrypt_byte(self, v):
        k = self.kdf()
        r = 0xFF & (v ^ k)
        self.acc = self.acc ^ r
        return r

    def decrypt_chunk(self, chunk):
        return bytes(map(self.decrypt_byte, chunk))


class QnapExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def _get_secret(self, header) -> str:
        raise NotImplementedError

    def extract(self, inpath: Path, outdir: Path) -> ExtractResult:
        fs = FileSystem(outdir)
        with (
            File.from_path(inpath) as infile,
            fs.open(Path(f"{inpath.name}.decrypted"), "wb+") as outfile,
        ):
            infile.seek(-FOOTER_LEN, io.SEEK_END)
            header = self._struct_parser.parse("qnap_header_t", infile, Endian.LITTLE)
            cryptor = Cryptor(self._get_secret(header))
            for chunk in iterate_file(infile, 0, header.encrypted_len, 1024):
                outfile.write(cryptor.decrypt_chunk(chunk))
            for chunk in iterate_file(
                infile,
                header.encrypted_len,
                infile.size() - FOOTER_LEN - header.encrypted_len,
                1024,
            ):
                outfile.write(chunk)
        return ExtractResult(reports=fs.problems)



import io

import attrs
from pyperscan import Flag, Pattern, Scan, StreamDatabase
from structlog import get_logger

from unblob.file_utils import File, stream_scan
from unblob.models import (
    Endian,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    StructParser,
    ValidChunk,
)

from ._qnap import C_DEFINITIONS, QnapExtractor

logger = get_logger()

SECRET = "QNAPNASVERSION"  # noqa: S105

FOOTER_PATTERN = [
    HexString("69 63 70 6e 61 73"),  # encrypted gzip stream start bytes
]

_STRUCT_PARSER = StructParser(C_DEFINITIONS)


@attrs.define
class QTSSearchContext:
    start_offset: int
    file: File
    end_offset: int


def is_valid_header(header) -> bool:
    try:
        header.device_id.decode("utf-8")
        header.file_version.decode("utf-8")
        header.firmware_date.decode("utf-8")
        header.revision.decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


def _hyperscan_match(
    context: QTSSearchContext, pattern_id: int, offset: int, end: int
) -> Scan:
    del pattern_id, end  # unused arguments
    if offset < context.start_offset:
        return Scan.Continue
    context.file.seek(offset, io.SEEK_SET)
    header = _STRUCT_PARSER.parse("qnap_header_t", context.file, Endian.LITTLE)
    logger.debug("qnap_header_t", header=header)

    if is_valid_header(header):
        context.end_offset = context.file.tell()
        return Scan.Terminate
    return Scan.Continue


def build_stream_end_scan_db(pattern_list):
    return StreamDatabase(
        *(Pattern(p.as_regex(), Flag.SOM_LEFTMOST, Flag.DOTALL) for p in pattern_list)
    )


hyperscan_stream_end_magic_db = build_stream_end_scan_db(FOOTER_PATTERN)


class QnapNasExtractor(QnapExtractor):
    def _get_secret(self, header) -> str:
        return SECRET + header.file_version.decode("utf-8")[0]


class QnapHandler(Handler):
    NAME = "qnap_nas"

    PATTERNS = [
        HexString("F5 7B 47 03"),
    ]
    EXTRACTOR = QnapNasExtractor()

    DOC = HandlerDoc(
        name="QNAP NAS",
        description="QNAP NAS firmware files consist of a custom header, encrypted data sections, and a footer marking the end of the encrypted stream. The header contains metadata such as device ID, firmware version, and encryption details.",
        handler_type=HandlerType.ARCHIVE,
        vendor="QNAP",
        references=[],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        context = QTSSearchContext(start_offset=start_offset, file=file, end_offset=-1)

        try:
            scanner = hyperscan_stream_end_magic_db.build(context, _hyperscan_match)  # type: ignore
            stream_scan(scanner, file)
        except Exception as e:
            logger.debug(
                "Error scanning for QNAP patterns",
                error=e,
            )
        if context.end_offset > 0:
            return ValidChunk(start_offset=start_offset, end_offset=context.end_offset)
        return None



import io
from pathlib import Path

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from structlog import get_logger

from unblob.file_utils import File, InvalidInputFormat
from unblob.models import (
    Endian,
    Extractor,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

C_DEFINITIONS = r"""
    typedef struct encrpted_img_header {
        char magic[12]; /* encrpted_img */
        uint32 file_size; /* total size of file */
    } dlink_header_t;
"""

HEADER_LEN = 16
PEB_SIZE = 0x20000
UBI_HEAD = b"\x55\x42\x49\x23\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
UBI_HEAD_LEN = len(UBI_HEAD)
KEY = b"he9-4+M!)d6=m~we1,q2a3d1n&2*Z^%8"
IV = b"J%1iQl8$=lm-;8AE"


class EncrptedExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        cipher = Cipher(algorithms.AES(KEY), modes.CBC(IV))
        decryptor = cipher.decryptor()
        outpath = outdir.joinpath(f"{inpath.name}.decrypted")
        outfile = outpath.open("wb")
        with inpath.open("rb") as f:
            f.seek(HEADER_LEN, io.SEEK_SET)  # skip header
            ciphertext = f.read(PEB_SIZE)
            while ciphertext and len(ciphertext) % 16 == 0:
                plaintext = decryptor.update(ciphertext)
                outfile.write(UBI_HEAD + plaintext[UBI_HEAD_LEN:])
                ciphertext = f.read(PEB_SIZE)
        outfile.write(decryptor.finalize())
        outfile.close()


class EncrptedHandler(StructHandler):
    NAME = "encrpted_img"

    PATTERNS = [
        Regex(r"encrpted_img"),
    ]
    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "dlink_header_t"
    EXTRACTOR = EncrptedExtractor()

    DOC = HandlerDoc(
        name="D-Link encrpted_img",
        description="A binary format used by D-Link to store encrypted firmware or data. It consists of a custom 12-byte magic header followed by the encrypted payload.",
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[
            Reference(
                title="How-To: Extracting Decryption Keys for D-Link",
                url="https://www.onekey.com/resource/extracting-decryption-keys-dlink",  # Replace with actual reference if available
            )
        ],
        limitations=[],
    )

    def is_valid_header(self, header) -> bool:
        return header.file_size >= len(header)

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, endian=Endian.BIG)

        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid header length")

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + len(header) + header.file_size,
        )



"""Searching Chunk related functions.

The main "entry point" is search_chunks_by_priority.
"""

import sys
from functools import lru_cache

import attrs
from pyperscan import Flag, Pattern, Scan, StreamDatabase
from structlog import get_logger

from .file_utils import DEFAULT_BUFSIZE, InvalidInputFormat, SeekError
from .handlers import Handlers
from .models import File, Handler, TaskResult, ValidChunk
from .parser import InvalidHexString
from .report import CalculateChunkExceptionReport

logger = get_logger()


@attrs.define
class HyperscanMatchContext:
    file: File
    file_size: int
    all_chunks: list
    task_result: TaskResult
    start_offset: int


def _calculate_chunk(
    handler: Handler, file: File, real_offset, task_result: TaskResult
) -> ValidChunk | None:
    file.seek(real_offset)
    try:
        return handler.calculate_chunk(file, real_offset)
    except InvalidInputFormat as exc:
        logger.debug(
            "File format is invalid",
            exc_info=exc,
            handler=handler.NAME,
            _verbosity=2,
        )
    except EOFError as exc:
        logger.debug(
            "File ends before header could be read",
            exc_info=exc,
            handler=handler.NAME,
            _verbosity=2,
        )
    except SeekError as exc:
        logger.debug(
            "Seek outside file during chunk calculation",
            exc_info=exc,
            handler=handler.NAME,
            _verbosity=2,
        )
    except Exception as exc:
        error_report = CalculateChunkExceptionReport(
            handler=handler.NAME,
            start_offset=real_offset,
            exception=exc,
        )
        task_result.add_report(error_report)
        logger.error(
            "Unhandled Exception during chunk calculation", **error_report.model_dump()
        )


def _hyperscan_match(
    context: HyperscanMatchContext, handler: Handler, offset: int, end: int
) -> Scan:
    del end  # unused argument
    offset += context.start_offset
    real_offset = offset + handler.PATTERN_MATCH_OFFSET

    # See https://github.com/vlaci/pyperscan/issues/110
    if real_offset < 0 or real_offset > sys.maxsize:
        return Scan.Continue

    # Skip chunk calculation if this would start inside another one,
    # similar to remove_inner_chunks, but before we even begin calculating.
    if any(chunk.contains_offset(real_offset) for chunk in context.all_chunks):
        logger.debug(
            "Skip chunk calculation as pattern is inside an other chunk",
            handler=handler.NAME,
            offset=real_offset,
            _verbosity=2,
        )
        return Scan.Continue

    logger.debug(
        "Calculating chunk for pattern match",
        start_offset=offset,
        real_offset=real_offset,
        _verbosity=2,
        handler=handler.NAME,
    )

    chunk = _calculate_chunk(handler, context.file, real_offset, context.task_result)

    # We found some random bytes this handler couldn't parse
    if chunk is None:
        return Scan.Continue

    if chunk.end_offset > context.file_size:
        logger.debug("Chunk overflows file", chunk=chunk, _verbosity=2)
        return Scan.Continue

    chunk.handler = handler
    logger.debug("Found valid chunk", chunk=chunk, handler=handler.NAME, _verbosity=1)
    context.all_chunks.append(chunk)
    context.start_offset = chunk.end_offset

    return Scan.Terminate


def stream_scan_chunks(scanner, file: File, context: HyperscanMatchContext):
    """Scan the whole file by increment of DEFAULT_BUFSIZE using Hyperscan's streaming mode."""
    i = context.start_offset
    with memoryview(file) as data:
        while i < file.size():
            if scanner.scan(data[i : i + DEFAULT_BUFSIZE]) == Scan.Terminate:
                scanner.reset()
                i = context.start_offset
            else:
                i += DEFAULT_BUFSIZE


def search_chunks(
    file: File,
    file_size: int,
    handlers: Handlers,
    task_result: TaskResult,
) -> list[ValidChunk]:
    """Search all ValidChunks within the file.

    Search for patterns and run Handler.calculate_chunk() on the found
    matches.  We don't deal with offset within already found
    ValidChunks and invalid chunks are thrown away.  If chunk covers
    the whole file we stop any further search and processing.
    """
    all_chunks = []

    hyperscan_db = build_hyperscan_database(handlers)

    hyperscan_context = HyperscanMatchContext(
        file=file,
        file_size=file_size,
        all_chunks=all_chunks,
        task_result=task_result,
        start_offset=0,
    )

    scanner = hyperscan_db.build(hyperscan_context, _hyperscan_match)  # type: ignore

    try:
        stream_scan_chunks(scanner, file, hyperscan_context)
    except Exception as e:
        logger.error(
            "Error scanning for patterns",
            error=e,
        )

    logger.debug(
        "Ended searching for chunks",
        all_chunks=all_chunks,
    )

    return all_chunks


@lru_cache
def build_hyperscan_database(handlers: Handlers) -> StreamDatabase:
    patterns = []
    for handler_class in handlers:
        handler = handler_class()
        for pattern in handler.PATTERNS:
            try:
                patterns.append(
                    Pattern(
                        pattern.as_regex(),
                        Flag.SOM_LEFTMOST,
                        Flag.DOTALL,
                        tag=handler,
                    )
                )
            except InvalidHexString as e:
                logger.error(
                    "Invalid pattern",
                    handler=handler.NAME,
                    pattern=pattern,
                    error=str(e),
                )
                raise
    return StreamDatabase(*patterns)



import abc
import atexit
import contextlib
import fcntl
import multiprocessing as mp
import multiprocessing.connection
import os
import queue
import signal
import sys
import threading
from collections.abc import Callable
from multiprocessing.queues import JoinableQueue, SimpleQueue
from typing import Any

from .logging import multiprocessing_breakpoint

mp.set_start_method("fork")


class PoolBase(abc.ABC):
    @abc.abstractmethod
    def submit(self, args):
        pass

    @abc.abstractmethod
    def process_until_done(self):
        pass

    def start(self):
        pass

    def close(self, *, immediate=False):
        pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, _exc_value, _tb):
        self.close(immediate=exc_type is not None)


class Queue(JoinableQueue):
    def is_empty(self) -> bool:
        """Check if all ``task_done`` has been called for all items.

        Based on ``multiprocessing.JoinableQueue.join``.
        """
        with self._cond:  # type: ignore
            return self._unfinished_tasks._semlock._is_zero()  # type: ignore  # noqa: SLF001


class ResultQueue(SimpleQueue):
    @property
    def reader(self) -> multiprocessing.connection.Connection:
        return self._reader  # type: ignore


class _Sentinel:
    pass


_SENTINEL = _Sentinel


class WorkerDiedError(RuntimeError):
    pass


def _worker_process(handler, input_, output, lifeline_worker_side, lifeline_host_side):
    # Creates a new process group, making sure no signals are
    # propagated from the main process to the worker processes.
    os.setpgrp()

    # Restore default signal handlers, otherwise workers would inherit
    # them from main process. When used as a library, the hosting app
    # is free to set-up its own signal handlers.
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    os.close(  # forked processes inherit open files, we don't need the host FD
        lifeline_host_side
    )

    def _exit_on_parent_death():
        os.read(lifeline_worker_side, 1)
        # We cannot really do anything about this, best to reliably
        # abort the process
        os._exit(1)

    parent_liveness_monitor = threading.Thread(
        target=_exit_on_parent_death, daemon=True
    )
    parent_liveness_monitor.start()

    sys.breakpointhook = multiprocessing_breakpoint
    while (args := input_.get()) is not _SENTINEL:
        result = handler(args)
        output.put(result)
    output.put(_SENTINEL)


class MultiPool(PoolBase):
    def __init__(
        self,
        process_num: int,
        handler: Callable[[Any], Any],
        *,
        result_callback: Callable[["MultiPool", Any], Any],
    ):
        if process_num <= 0:
            raise ValueError("At process_num must be greater than 0")

        self._running = False
        self._result_callback = result_callback
        self._input = Queue(ctx=mp.get_context())
        self._input.cancel_join_thread()
        self._output = ResultQueue(ctx=mp.get_context())
        # see search results for "death-pipe" or "forkfd concept"
        (self._lifeline_worker_side, self._lifeline_host_side) = os.pipe()
        fcntl.fcntl(self._lifeline_host_side, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
        fcntl.fcntl(self._lifeline_worker_side, fcntl.F_SETFD, fcntl.FD_CLOEXEC)

        self._procs = [
            mp.Process(
                target=_worker_process,
                args=(
                    handler,
                    self._input,
                    self._output,
                    self._lifeline_worker_side,
                    self._lifeline_host_side,
                ),
            )
            for _ in range(process_num)
        ]
        self._tid = threading.get_native_id()

    def start(self):
        self._running = True
        for p in self._procs:
            p.start()
        # We are the host process, we don't need this anymore.
        # Had to keep the file alive until inherited by the forked subprocess
        os.close(self._lifeline_worker_side)
        atexit.register(self._close_immediate)

    def _any_worker_exited(self) -> bool:
        sentinels = [p.sentinel for p in self._procs]
        return bool(multiprocessing.connection.wait(sentinels, timeout=0))

    def close(self, *, immediate=False):
        if not self._running:
            return
        self._running = False
        atexit.unregister(self._close_immediate)
        immediate = immediate or self._any_worker_exited()

        termination_exception = None
        if not immediate:
            try:
                self._clear_input_queue()
                self._request_workers_to_quit()
                self._clear_output_queue()
            except BaseException as exc:
                termination_exception = exc
                immediate = True

        if immediate:
            self._terminate_workers()

        self._wait_for_workers_to_quit()

        # closing this FD any sooner would cause workers to abort
        # immediately, should close only after workers quit
        os.close(self._lifeline_host_side)

        if termination_exception:
            raise termination_exception

    def _close_immediate(self):
        self.close(immediate=True)

    def _terminate_workers(self):
        for proc in self._procs:
            proc.terminate()

        self._input.close()
        self._output.close()

    def _clear_input_queue(self):
        with contextlib.suppress(queue.Empty):
            while True:
                self._input.get_nowait()

    def _request_workers_to_quit(self):
        for proc in self._procs:
            if proc.exitcode is not None:
                continue
            self._input.put(_SENTINEL)
        self._input.close()

    def _clear_output_queue(self):
        alive = {p.sentinel: p for p in self._procs if p.exitcode is None}
        while alive:
            ready = multiprocessing.connection.wait([self._output.reader, *alive])
            for fd in ready:
                alive.pop(fd, None)  # type: ignore[arg-type]
            if self._output.reader in ready:
                self._output.get()

    def _wait_for_workers_to_quit(self):
        for p in self._procs:
            p.join()

    def submit(self, args):
        if threading.get_native_id() != self._tid:
            raise RuntimeError(
                "Submit can only be called from the same "
                "thread/process where the pool is created"
            )
        self._input.put(args)

    def _check_worker_deaths(self, sentinels, ready):
        for fd in ready:
            if fd not in sentinels:
                continue
            proc = sentinels.pop(fd)
            if proc.exitcode != 0:
                exitcode = proc.exitcode
                if exitcode is not None and exitcode < 0:
                    reason = f"killed by signal {-exitcode}"
                else:
                    reason = f"exited with code {exitcode}"
                raise WorkerDiedError(
                    f"Worker process {proc.pid} exited unexpectedly ({reason})"
                )

    def process_until_done(self):
        sentinels = {p.sentinel: p for p in self._procs}
        with contextlib.suppress(EOFError):
            while not self._input.is_empty():
                ready = multiprocessing.connection.wait(
                    [self._output.reader, *sentinels]
                )
                self._check_worker_deaths(sentinels, ready)
                if self._output.reader in ready:
                    result = self._output.get()
                    self._result_callback(self, result)
                    self._input.task_done()


class SinglePool(PoolBase):
    def __init__(self, handler, *, result_callback):
        self._handler = handler
        self._result_callback = result_callback

    def submit(self, args):
        result = self._handler(args)
        self._result_callback(self, result)

    def process_until_done(self):
        pass


def make_pool(process_num, handler, result_callback) -> SinglePool | MultiPool:
    if process_num == 1:
        return SinglePool(handler=handler, result_callback=result_callback)

    return MultiPool(
        process_num=process_num,
        handler=handler,
        result_callback=result_callback,
    )



from __future__ import annotations

import io
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import (
    Cipher,
    algorithms,
    modes,
)

from unblob._rust import math_tools
from unblob.file_utils import Endian, File, FileSystem, InvalidInputFormat, StructParser
from unblob.models import (
    Extractor,
    ExtractResult,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from cryptography.hazmat.primitives.ciphers import CipherContext

C_DEFINITIONS = r"""
    typedef struct dlink_alpha_header {
        uint32 magic1;
        uint32 signature_len;
        uint32 padding;
        char   signature[signature_len];    /* "signature=XXXXXX_dlink.YYYY_ZZZZZZ\0" */
        // where XXXXXX seems to be the Alpha Networks model number, YYYY the year
        // and ZZZZZZ the D-Link model number
        uint32 magic2;
        uint32 signature_len2;
        uint32 fsize;                       /* size of the firmware image (without 128 Bytes header) */
    } dlink_alpha_header_t;

    // see https://github.com/openwrt/firmware-utils/blob/master/src/mkwrggimg.c
    typedef struct wrgg03_header {
        char        signature[32];    /* signature without "signature=" */
        uint32      magic1;
        uint32      magic2;
        char        version[16];
        char        model[16];
        uint32      flag[2];
        uint32  	reserve[2];
        char        buildno[16];
        uint32  	fsize;
        uint32  	offset;
        char		devname[32];
        char		digest[16];
    } wrgg03_header_t;
"""
XOR_RANGE = 0xFC
UNENCRYPTED_ENTROPY_THRESHOLD = 3


@dataclass
class EncParams:
    signature: bytes
    key: bytes
    iv: bytes

    @property
    def mangled_key(self):
        return self._mangle(self.signature, self.key)

    @property
    def mangled_iv(self):
        return self._mangle(self.signature, self.iv)

    @staticmethod
    def _mangle(image_sign: bytes, data: bytes) -> bytes:
        sign_len = len(image_sign)
        return bytes(
            data_byte ^ ((i + 1) % XOR_RANGE) ^ image_sign[i % sign_len]
            for i, data_byte in enumerate(data)
        )


@dataclass
class EncParamsV2(EncParams):
    key: bytes = b"oVhq0hvXHdfaGFLdubM4/QvuVHdKee7v"
    iv: bytes = b"0BO5nlYankuVBe4s"


ENC_START_TO_PARAMS = {
    # DAP-1665 B
    bytes.fromhex("35 66 6f 68"): EncParams(
        signature=b"wapac25_dlink.2015_dap1665",
        key=b"EfCHXytwsC6F0zsedwZc+9vDbCjE3ge4",
        iv=b"ggPy917jwESpnfXm",
    ),
    # DAP-1720 A
    bytes.fromhex("68 01 cc fb"): EncParams(
        signature=b"wapac28_dlink.2015_dap1720",
        key=b"qBiz6o/1RVQTtJBd3FS7FDbqogE8yoBm",
        iv=b"EfDMqWWxHCOhEqgY",
    ),
    # DIR-822 C1
    bytes.fromhex("df 8c 39 0d"): EncParams(
        signature=b"wrgac43s_dlink.2015_dir822c1",
        key=b"KNpsEntCcsep1jdFIs3wnXySKRGNCGmf",
        iv=b"uph587JdKHrtAUlr",
    ),
    # DIR-842 C1 / C2
    bytes.fromhex("f5 2a a0 b4"): EncParams(
        signature=b"wrgac65_dlink.2015_dir842",
        key=b"xQYoRZeD726UAbRb846kO7TeNw8eZa6u",
        iv=b"zufEbNF3kUafxFiE",
    ),
    # DIR-842 C3 (same as C1 except for the "EU" at the end of the signature)
    bytes.fromhex("21 dd da 00"): EncParams(
        signature=b"wrgac65_dlink.2015_dir842EU",
        key=b"xQYoRZeD726UAbRb846kO7TeNw8eZa6u",
        iv=b"zufEbNF3kUafxFiE",
    ),
    # DIR-850L A1
    bytes.fromhex("e3 13 00 5b"): EncParams(
        signature=b"wrgac05_dlob.hans_dir850l",
        key=b"BIuS1CVMEQG+0pUeE99jnR+vLlLd9unr",
        iv=b"f3+odwHhmJL1ceW1",
    ),
    # DIR-850L B1
    bytes.fromhex("0a 14 e4 24"): EncParams(
        signature=b"wrgac25_dlink.2013gui_dir850l",
        key=b"qQehHMEmEPQ5izL+cabn8bNHZXHjkp6W",
        iv=b"Mmb+IKQgnO8OuF4b",
    ),
    # DIR-859 A
    bytes.fromhex("4c 1b 95 af"): EncParams(
        signature=b"wrgac37_dlink.2013gui_dir859",
        key=b"KY0H9R2PDL3eu1J4uCVd1CK7BJ7vF1kc",
        iv=b"qbStAzIRvWeQHz5U",
    ),
    # FixMe: add more entries when they are found
}


def decrypt_chunk(encrypted: bytes, enc_params: EncParams) -> bytes:
    decryptor = _get_decryptor(enc_params)
    return decryptor.update(encrypted) + decryptor.finalize()


def decrypt_file(
    file: File, enc_params: EncParams, chunk_size: int = 16
) -> Iterable[bytes]:
    decryptor = _get_decryptor(enc_params)
    while chunk := file.read(chunk_size):
        yield decryptor.update(chunk)
    yield decryptor.finalize()


def _get_decryptor(enc_params: EncParams) -> CipherContext:
    cipher = Cipher(
        algorithms.AES(enc_params.mangled_key),
        modes.CBC(enc_params.mangled_iv),
        backend=default_backend(),
    )
    return cipher.decryptor()


class AlphaEncimgExtractorV1(Extractor):
    def extract(self, inpath: Path, outdir: Path) -> ExtractResult | None:
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            enc_magic = file.read(4)
            enc_params = ENC_START_TO_PARAMS.get(enc_magic)
            if not enc_params:
                raise InvalidInputFormat("Device not supported")
            file.seek(0, io.SEEK_SET)
            out_path = Path(f"{enc_params.signature.decode()}.bin")
            fs.write_chunks(out_path, decrypt_file(file, enc_params))
        return ExtractResult(reports=fs.problems)


class AlphaEncimgExtractorV2(Extractor):
    def extract(self, inpath: Path, outdir: Path) -> ExtractResult | None:
        struct_parser = StructParser(C_DEFINITIONS)
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            header = struct_parser.parse("wrgg03_header_t", file, Endian.BIG)
            enc_params = EncParamsV2(signature=header.signature.rstrip(b"\0"))
            file.seek(AlphaEncimgV2Handler.HEADER_SIZE, io.SEEK_SET)
            out_path = Path(f"{enc_params.signature.decode()}.bin")
            fs.write_chunks(out_path, decrypt_file(file, enc_params))
        return ExtractResult(reports=fs.problems)


class AlphaEncimgHandler(StructHandler):
    NAME = "alpha_encimg_v1"
    PATTERNS = [
        HexString("35 66 6f 68 ef 1a fe 1f 34 ef 4f 11 21 05 4e be"),  # DAP-1665 B
        HexString("68 01 cc fb ad 6b a0 ba 33 04 b0 9c bb 48 b0 27"),  # DAP-1720 A
        HexString("df 8c 39 0d 22 b4 dc 29 fb 4e bf db e8 e1 8b fb"),  # DIR-822 C1
        HexString("f5 2a a0 b4 92 53 bf ef f8 21 a6 2e 28 a7 39 8b"),  # DIR-842 C1 / C2
        HexString("21 dd da 00 99 d8 87 a9 d5 2d 7e ff 3b 58 70 a6"),  # DIR-842 C3
        HexString("e3 13 00 5b 76 df 0b e8 83 24 5a 42 ff 91 2d 3d"),  # DIR-850L A1
        HexString("0a 14 e4 24 ff 0f b4 d7 53 66 a0 b0 72 fe ab df"),  # DIR-850L B1
        HexString("4c 1b 95 af 93 72 5f 81 03 03 96 4d dd 76 01 74"),  # DIR-859 A
        HexString("98 9c 3c 2c 6b 19 1b 25 6e 03 ee 0f 72 5c 6c a0"),  # DIR-865L A
        HexString("3d e6 02 63 3f c1 85 e7 d2 bf 64 af 29 57 3a bd"),  # DIR-868L A
        HexString("8e 69 57 e7 76 09 d7 94 47 75 78 a2 4a 1f c9 b2"),  # DIR-880L A
        HexString("92 61 58 58 14 db bb 3b e5 a5 f3 e7 10 9c a2 0b"),  # DIR-885L A
        HexString("cb c3 a7 4c 50 0f 42 43 a5 d9 7c a5 25 6b cd ba"),  # DIR-890L A
        HexString("91 5a c9 a2 11 ff aa 6d b0 12 e8 8d 2c 3c 23 cb"),  # DIR-895L A
    ]
    EXTRACTOR = AlphaEncimgExtractorV1()
    C_DEFINITIONS = C_DEFINITIONS
    DOC = HandlerDoc(
        name="D-Link Alpha encimg V1 Firmware",
        description=(
            "Encrypted firmware images found in D-Link DIR devices manufactured by Alpha Networks."
            "Uses AES-256-CBC encryption with device-specific keys."
        ),
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[
            Reference(
                title="OpenWRT forum",
                url="https://forum.openwrt.org/t/adding-openwrt-support-for-d-link-dir-x1860-mt7621-mt7915-ax1800/106500",
            ),
            Reference(
                title="delink tool",
                url="https://github.com/devttys0/delink/blob/main/src/encimg.rs",
            ),
        ],
        limitations=[],
    )
    HEADER_STRUCT = "dlink_alpha_header_t"
    HEADER_SIZE = 0x80
    _SIGNATURE_PREFIX = b"signature="
    MAGIC = {0x5EA3A417}

    def _get_header(self, file: File):
        encrypted_header = file.read(self.HEADER_SIZE)
        enc_params = ENC_START_TO_PARAMS.get(encrypted_header[:4])
        if not enc_params:
            raise InvalidInputFormat("Device not supported")

        decrypted_header = decrypt_chunk(encrypted_header, enc_params)
        header = self._struct_parser.parse(
            self.HEADER_STRUCT, decrypted_header, Endian.BIG
        )
        self._validate(enc_params, file, header)
        return header

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self._get_header(file)
        end_offset = start_offset + self.HEADER_SIZE + header.fsize
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)

    def _validate(self, enc_params: EncParams, file: File, header):
        if header.magic1 not in self.MAGIC:
            raise InvalidInputFormat(f"Invalid magic: {header.magic1}")

        expected_signature = self._SIGNATURE_PREFIX + enc_params.signature
        if header.signature.rstrip(b"\0") != expected_signature:
            raise InvalidInputFormat(f"Invalid signature {header.signature}")

        remaining = file.size() - file.tell()
        if header.fsize > remaining:
            raise InvalidInputFormat("Invalid file size")

        if header.fsize % 16:
            raise InvalidInputFormat(
                f"Firmware size not aligned to 16 bytes: {header.fsize}"
            )


class AlphaEncimgV2Handler(AlphaEncimgHandler):
    NAME = "alpha_encimg_v2"
    PATTERNS = [
        HexString("77 61 70 [29] 21 03 08 20 21 03 08 20"),  # Alpha Networks DAP-2xxx
    ]
    EXTRACTOR = AlphaEncimgExtractorV2()
    DOC = HandlerDoc(
        name="D-Link Alpha encimg v2 Firmware",
        description=(
            "Encrypted firmware images found in D-Link DIR devices manufactured by Alpha Networks."
            "Uses AES-256-CBC encryption with device-specific keys."
            "Unlike the other variant, this one uses a prepended unencrypted WRGG03 header."
        ),
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[
            Reference(
                title="OpenWRT Wiki",
                url="https://openwrt.org/toh/d-link/d-link_dap_series_of_business_access_points#old_generation_dap-2xxxdap-3xxx_built_by_alpha_networks",
            ),
            Reference(
                title="delink tool",
                url="https://github.com/devttys0/delink/blob/main/src/encimg.rs",
            ),
        ],
        limitations=[],
    )
    HEADER_SIZE = 0xA0
    HEADER_STRUCT = "wrgg03_header_t"
    _SIGNATURE_PREFIX = b""
    MAGIC = {0x20080321, 0x21030820}
    KNOWN_UNENCRYPTED = (
        bytes.fromhex("5d 00 00 80"),
        bytes.fromhex("d0 0d fe ed"),
        bytes.fromhex("10 ec 7a 0d"),  # DAP X2810/X2850
    )

    def _validate(self, enc_params: EncParams, file: File, header):
        first_block = file.read(16)
        file.seek(-16, os.SEEK_CUR)
        if self._is_not_encrypted(first_block):
            raise InvalidInputFormat("Firmware payload does not seem to be encrypted")
        super()._validate(enc_params, file, header)

    def _is_not_encrypted(self, first_block: bytes) -> bool:
        return (
            any(first_block.startswith(m) for m in self.KNOWN_UNENCRYPTED)
            or math_tools.shannon_entropy(first_block) < UNENCRYPTED_ENTROPY_THRESHOLD
        )

    def _get_header(self, file: File):
        current_offset = file.tell()
        header_le = self.parse_header(file, Endian.LITTLE)
        file.seek(current_offset, io.SEEK_SET)
        header_be = self.parse_header(file, Endian.BIG)
        # there are little and big endian devices that use this format, but there seems to be
        # only one mandatory integer header field: size (sadly we cannot write signatures for that)
        # since the size is an uint32 and the FW image should be smaller than 64 MB, it is safe to assume
        # that the correct endianness is the one with the smaller size field
        header = header_le if header_le.fsize < header_be.fsize else header_be

        device = header.signature.rstrip(b"\0")
        enc_params = EncParamsV2(signature=device)
        self._validate(enc_params, file, header)
        return header



import pluggy

from unblob.models import DirectoryHandler, Handler
from unblob.report import Report

hookspec = pluggy.HookspecMarker("unblob")


@hookspec
def unblob_register_handlers() -> list[type[Handler]]:
    """Register handler types to known handlers.

    :returns: The list of handlers to be registered
    """
    return []


@hookspec
def unblob_register_dir_handlers() -> list[type[DirectoryHandler]]:
    """Register directory handler types to known handlers.

    :returns: The list of directory handlers to be registered
    """
    return []


@hookspec
def unblob_register_reports() -> list[type[Report]]:
    """Register report types to known reports.

    :returns: The list of report types to be registered
    """
    return []



import binascii
import io
import itertools
from pathlib import Path
from typing import TYPE_CHECKING, cast

from structlog import get_logger

from unblob.extractor import carve_chunk_to_file
from unblob.file_utils import Endian, File, InvalidInputFormat, StructParser
from unblob.models import (
    Chunk,
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    StructHandler,
    ValidChunk,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

logger = get_logger()
CRC_CONTENT_OFFSET = 12

TRX_V1_C_DEFINITION = r"""
    typedef struct trx_header {
        uint32 magic;              /* "HDR0" */
        uint32 len;                /* header size + data */
        uint32 crc32;              /* 32-bit CRC from flag_version to end of file */
        uint16 flags;
        uint16 version;
        uint32 offsets[3];         /* Offsets of partitions from start of header */
    } trx_header_t;
"""

TRX_V2_C_DEFINITION = r"""
    typedef struct trx_header {
        uint32 magic;              /* "HDR0" */
        uint32 len;                /* header size + data */
        uint32 crc32;              /* 32-bit CRC from flag_version to end of file */
        uint16 flags;
        uint16 version;
        uint32 offsets[4];         /* Offsets of partitions from start of header */
    } trx_header_t;
"""


class TRXExtractor(Extractor):
    def __init__(self, c_definitions: str):
        self._struct_parser = StructParser(c_definitions)

    def extract(self, inpath: Path, outdir: Path):
        with File.from_path(inpath) as file:
            header = self._struct_parser.parse("trx_header_t", file, Endian.LITTLE)
            file.seek(0, io.SEEK_END)
            eof = file.tell()
            offsets = sorted(
                [
                    offset
                    for offset in [*cast("Iterable", header.offsets), eof]
                    if offset > 0
                ]
            )
            for i, (start_offset, end_offset) in enumerate(itertools.pairwise(offsets)):
                chunk = Chunk(start_offset=start_offset, end_offset=end_offset)
                carve_chunk_to_file(outdir.joinpath(Path(f"part{i}")), file, chunk)


class NetgearTRXBase(StructHandler):
    HEADER_STRUCT = "trx_header_t"

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, endian=Endian.LITTLE)

        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid TRX header.")

        if not self._is_crc_valid(file, start_offset, header):
            raise InvalidInputFormat("Invalid CRC32.")

        return ValidChunk(
            start_offset=start_offset, end_offset=start_offset + header.len
        )

    def is_valid_header(self, header) -> bool:
        return header.len >= len(header)

    def _is_crc_valid(self, file: File, start_offset: int, header) -> bool:
        file.seek(start_offset + CRC_CONTENT_OFFSET)
        content = bytearray(file.read(header.len - CRC_CONTENT_OFFSET))
        computed_crc = (binascii.crc32(content) ^ -1) & 0xFFFFFFFF
        return header.crc32 == computed_crc


class NetgearTRXv1Handler(NetgearTRXBase):
    NAME = "trx_v1"
    PATTERNS = [
        HexString(
            """
            // 00000000: 4844 5230 0010 0000 33b9 d625 0000 0100  HDR0....3..%....
            // 00000010: 1c00 0000 3000 0000 4400 0000 7361 6d70  ....0...D...
            48 44 52 30 [10] 01 00
        """
        ),
    ]
    C_DEFINITIONS = TRX_V1_C_DEFINITION
    EXTRACTOR = TRXExtractor(TRX_V1_C_DEFINITION)

    DOC = HandlerDoc(
        name="Netgear TRX v1",
        description="Netgear TRX v1 firmware format includes a custom header with partition offsets and a CRC32 checksum for integrity verification. It supports up to three partitions defined in the header.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Netgear",
        references=[],
        limitations=[],
    )


class NetgearTRXv2Handler(NetgearTRXBase):
    NAME = "trx_v2"
    PATTERNS = [
        HexString(
            """
            // 00000000: 4844 5230 0010 0000 0a3d 4b05 0000 0200  HDR0.....=K.....
            // 00000010: 2000 0000 3400 ffff ffff ffff ffff 0000   ...4...........
            48 44 52 30 [10] 02 00
        """
        ),
    ]
    C_DEFINITIONS = TRX_V2_C_DEFINITION
    EXTRACTOR = TRXExtractor(TRX_V2_C_DEFINITION)

    DOC = HandlerDoc(
        name="Netgear TRX v2",
        description="Netgear TRX v2 firmware format includes a custom header with partition offsets and a CRC32 checksum for integrity verification. It supports up to four partitions defined in the header.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Netgear",
        references=[],
        limitations=[],
    )



import io

from unblob.extractors import Command
from unblob.file_utils import (
    Endian,
    InvalidInputFormat,
    snull,
)
from unblob.models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

C_DEFINITIONS = r"""
    typedef struct erofs_handler{
        uint32_t magic;
        uint32_t crc32c;
        uint32_t feature_compact;
        uint8_t block_size_bs;
        uint8_t sb_extslots;
        uint16_t root_nid;
        uint64_t inos;
        uint64_t build_time;
        uint32_t build_time_nsec;
        uint32_t block_count;
        uint32_t meta_blkaddr;
        uint32_t xattr_blkaddr;
        uint8_t uuid[16];
        char volume_name[16];
        uint32_t feature_incompact;
        char reserved[44];
    } erofs_handler_t;
"""

SUPERBLOCK_OFFSET = 0x400


class EROFSHandler(StructHandler):
    NAME = "erofs"
    PATTERNS = [HexString("e2 e1 f5 e0")]  # Magic in little endian
    HEADER_STRUCT = "erofs_handler_t"
    C_DEFINITIONS = C_DEFINITIONS
    EXTRACTOR = Command(
        "fsck.erofs",
        "--no-preserve",
        "--extract={outdir}",
        "{inpath}",
    )
    PATTERN_MATCH_OFFSET = -SUPERBLOCK_OFFSET

    DOC = HandlerDoc(
        name="Android EROFS",
        description="EROFS (Enhanced Read-Only File System) is a lightweight, high-performance file system designed for read-only use cases, commonly used in Android and Linux. It features compression support, metadata efficiency, and a fixed superblock structure.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="Google",
        references=[
            Reference(
                title="EROFS Documentation",
                url="https://www.kernel.org/doc/html/latest/filesystems/erofs.html",
            ),
            Reference(
                title="EROFS Wikipedia",
                url="https://en.wikipedia.org/wiki/Enhanced_Read-Only_File_System",
            ),
        ],
        limitations=[],
    )

    def is_valid_header(self, header) -> bool:
        try:
            snull(header.volume_name).decode("utf-8")
        except UnicodeDecodeError:
            return False
        return (
            header.block_count >= 1
            and header.build_time > 0
            and header.build_time_nsec > 0
            and header.block_size_bs >= 9
        )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset + SUPERBLOCK_OFFSET, io.SEEK_SET)
        header = self.parse_header(file, Endian.LITTLE)
        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid erofs header.")

        end_offset = (1 << header.block_size_bs) * header.block_count
        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



import io
import struct

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import InvalidInputFormat
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

FREE_SECTOR = 0xFFFFFFFF
END_OF_CHAIN = 0xFFFFFFFE
HEADER_SIZE = 512

logger = get_logger()


class MsiHandler(StructHandler):
    NAME = "msi"

    PATTERNS = [HexString("D0 CF 11 E0 A1 B1 1A E1")]
    C_DEFINITIONS = r"""
        typedef struct cfbf_header
        {
                                        // [offset from start (bytes), length (bytes)]
            uint8 signature[8];         // [00H,08] {0xd0, 0xcf, 0x11, 0xe0, 0xa1, 0xb1,
                                        // 0x1a, 0xe1} for current version
            uint8 clsid[16];            // [08H,16] reserved must be zero (WriteClassStg/
                                        // GetClassFile uses root directory class id)
            uint16 minorVersion;        // [18H,02] minor version of the format: 33 is
                                        // written by reference implementation
            uint16 dllVersion;          // [1AH,02] major version of the dll/format: 3 for
                                        // 512-byte sectors, 4 for 4 KB sectors
            uint16 byteOrder;           // [1CH,02] 0xFFFE: indicates Intel byte-ordering
            uint16 sectorShift;         // [1EH,02] size of sectors in power-of-two;
                                        // typically 9 indicating 512-byte sectors
            uint16 miniSectorShift;     // [20H,02] size of mini-sectors in power-of-two;
                                        // typically 6 indicating 64-byte mini-sectors
            uint16 reserved;            // [22H,02] reserved, must be zero
            uint32 reserved1;           // [24H,04] reserved, must be zero
            uint32 csectDir;            // [28H,04] must be zero for 512-byte sectors,
                                        // number of SECTs in directory chain for 4 KB
                                        // sectors
            uint32 csectFat;            // [2CH,04] number of SECTs in the FAT chain
            uint32 sectDirStart;        // [30H,04] first SECT in the directory chain
            uint32 txSignature;         // [34H,04] signature used for transactions; must
                                        // be zero. The reference implementation
                                        // does not support transactions
            uint32 miniSectorCutoff;    // [38H,04] maximum size for a mini stream;
                                        // typically 4096 bytes
            uint32 sectMiniFatStart;    // [3CH,04] first SECT in the MiniFAT chain
            uint32 csectMiniFat;        // [40H,04] number of SECTs in the MiniFAT chain
            uint32 sectDifStart;        // [44H,04] first SECT in the DIFAT chain
            uint32 csectDif;            // [48H,04] number of SECTs in the DIFAT chain
            uint32 sectFat[109];        // [4CH,436] the SECTs of first 109 FAT sectors
         } cfbf_header_t;
    """
    HEADER_STRUCT = "cfbf_header_t"

    EXTRACTOR = Command("7z", "x", "-p", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="MSI",
        description="Microsoft Installer (MSI) files are used for the installation, maintenance, and removal of software.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Microsoft",
        references=[
            Reference(
                title="MSI File Format Documentation",
                url="https://docs.microsoft.com/en-us/windows/win32/msi/overview-of-windows-installer",
            ),
            Reference(
                title="Compound File Binary Format",
                url="https://en.wikipedia.org/wiki/Compound_File_Binary_Format",
            ),
        ],
        limitations=[
            "Limited to CFB based extraction, not full-on MSI extraction",
            "Extracted files have names coming from CFB internal representation, and may not correspond to the one they would have on disk after running the installer",
        ],
    )

    def _read_sector(
        self, file: File, start_offset: int, sector_size: int, sector_id: int
    ) -> bytes:
        # All sectors, including the fixed-size header, occupy full sector_size
        sector_offset = start_offset + sector_size + sector_id * sector_size
        if sector_offset > file.size():
            raise InvalidInputFormat("Invalid MSI file, sector offset too large")

        file.seek(sector_offset, io.SEEK_SET)
        raw_sector = file.read(sector_size)
        if len(raw_sector) != sector_size:
            raise InvalidInputFormat("Invalid MSI file, sector shorter than expected")

        return raw_sector

    def _append_fat_sector(
        self, fat_sectors: list[int], sector_id: int, required_count: int
    ) -> bool:
        if sector_id == FREE_SECTOR:
            return False

        fat_sectors.append(sector_id)
        return len(fat_sectors) >= required_count

    def _extend_fat_from_difat(
        self,
        file: File,
        header,
        start_offset: int,
        sector_size: int,
        entries_per_sector: int,
        fat_sectors: list[int],
    ) -> None:
        difat_sector = header.sectDifStart

        for _ in range(header.csectDif):
            if difat_sector in (FREE_SECTOR, END_OF_CHAIN):
                break

            raw_sector = self._read_sector(
                file, start_offset, sector_size, difat_sector
            )
            entries = struct.unpack(f"<{entries_per_sector}I", raw_sector)

            difat_sector = entries[-1]
            for entry in entries[:-1]:
                if self._append_fat_sector(
                    fat_sectors, entry, required_count=header.csectFat
                ):
                    return

    def _collect_fat_sectors(
        self,
        file: File,
        header,
        start_offset: int,
        sector_size: int,
        entries_per_sector: int,
    ) -> list[int]:
        fat_sectors: list[int] = []

        for sect in header.sectFat:
            if self._append_fat_sector(fat_sectors, sect, header.csectFat):
                return fat_sectors

        if len(fat_sectors) < header.csectFat:
            self._extend_fat_from_difat(
                file, header, start_offset, sector_size, entries_per_sector, fat_sectors
            )

        if len(fat_sectors) != header.csectFat:
            raise InvalidInputFormat("Invalid MSI file, incomplete FAT chain")

        return fat_sectors

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset, io.SEEK_SET)
        header = self.parse_header(file)

        sector_size = 2**header.sectorShift
        entries_per_sector = sector_size // 4

        if sector_size < HEADER_SIZE:
            raise InvalidInputFormat("Invalid MSI file, sector smaller than header")

        if header.csectFat == 0:
            raise InvalidInputFormat("Invalid MSI file, FAT chain is empty")

        fat_sectors = self._collect_fat_sectors(
            file, header, start_offset, sector_size, entries_per_sector
        )

        max_used_sector = 0
        for sector_index, sect in enumerate(fat_sectors):
            raw_sector = self._read_sector(file, start_offset, sector_size, sect)
            entries = struct.unpack(f"<{entries_per_sector}I", raw_sector)

            base_sector_id = sector_index * entries_per_sector
            for entry_id in range(len(entries) - 1, -1, -1):
                if entries[entry_id] == FREE_SECTOR:
                    continue

                max_id = base_sector_id + entry_id
                max_used_sector = max(max_used_sector, max_id)
                break

        total_size = sector_size + ((max_used_sector + 1) * sector_size)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + total_size,
        )



import io
from pathlib import Path

from structlog import get_logger

from unblob.file_utils import (
    Endian,
    File,
    FileSystem,
    InvalidInputFormat,
    StructParser,
    snull,
)
from unblob.models import (
    Extractor,
    ExtractResult,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

C_DEFINITIONS = r"""
    typedef struct ipkg_file_entry {
        char name[256];
        uint64 offset;
        uint64 entry_size;
        uint32 crc32;
    } ipkg_toc_entry_t;

    typedef struct ipkg_header {
        char magic[4];
        uint16 major;
        uint16 minor;
        uint32 toc_offset;
        uint32 unknown_1;
        uint32 toc_entries;
        uint32 unknown_2[2];
        uint32 always_null;
        char file_version[256];
        char product_name[256];
        char ipkg_name[256];
        char signature[256];
    } ipkg_header_t;
"""


def is_valid_header(header) -> bool:
    if header.toc_offset == 0 or header.toc_entries == 0:
        return False
    try:
        snull(header.ipkg_name).decode("utf-8")
        snull(header.file_version).decode("utf-8")
        snull(header.product_name).decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


class HPIPKGExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        entries = []
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            header = self._struct_parser.parse("ipkg_header_t", file, Endian.LITTLE)
            file.seek(header.toc_offset, io.SEEK_SET)
            for _ in range(header.toc_entries):
                entry = self._struct_parser.parse(
                    "ipkg_toc_entry_t", file, Endian.LITTLE
                )
                entry_path = Path(snull(entry.name).decode("utf-8"))
                if entry_path.parent.name:
                    raise InvalidInputFormat("Entry name contains directories.")
                entries.append(
                    (
                        Path(entry_path.name),
                        entry.offset,
                        entry.entry_size,
                    )
                )

            for carve_path, start_offset, size in entries:
                fs.carve(carve_path, file, start_offset, size)

            return ExtractResult(reports=fs.problems)


class HPIPKGHandler(StructHandler):
    NAME = "ipkg"

    PATTERNS = [HexString("69 70 6B 67 01 00 03 00")]

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "ipkg_header_t"
    EXTRACTOR = HPIPKGExtractor()

    DOC = HandlerDoc(
        name="HP IPKG",
        description="HP IPKG firmware archives consist of a custom header, followed by a table of contents and file entries. Each entry specifies metadata such as file name, offset, size, and CRC32 checksum.",
        handler_type=HandlerType.ARCHIVE,
        vendor="HP",
        references=[
            Reference(
                title="hpbdl",
                url="https://github.com/tylerwhall/hpbdl",
            )
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file, endian=Endian.LITTLE)

        if not is_valid_header(header):
            raise InvalidInputFormat("Invalid IPKG header.")

        file.seek(start_offset + header.toc_offset, io.SEEK_SET)
        end_offset = -1
        for _ in range(header.toc_entries):
            entry = self._struct_parser.parse("ipkg_toc_entry_t", file, Endian.LITTLE)
            end_offset = max(end_offset, start_offset + entry.offset + entry.entry_size)

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



from __future__ import annotations

import io
import os
import stat
import struct
from enum import IntEnum, unique
from pathlib import Path

from structlog import get_logger

from ...file_utils import (
    Endian,
    FileSystem,
    InvalidInputFormat,
    read_until_past,
    round_up,
)
from ...models import (
    Extractor,
    ExtractResult,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


STRING_ALIGNMENT = 16
MAX_LINUX_PATH_LENGTH = 0xFF
MAX_UINT32 = 0x100000000


WORLD_RW = 0o666
WORLD_RWX = 0o777
ROMFS_HEADER_SIZE = 512
ROMFS_SIGNATURE = b"-rom1fs-"


@unique
class FSType(IntEnum):
    HARD_LINK = 0
    DIRECTORY = 1
    FILE = 2
    SYMLINK = 3
    BLOCK_DEV = 4
    CHAR_DEV = 5
    SOCKET = 6
    FIFO = 7


def valid_checksum(content: bytes) -> bool:
    """Apply a RomFS checksum and returns whether it's valid or not."""
    total = 0

    # unalign content will lead to unpacking errors down the line
    if len(content) % 4 != 0:
        return False

    for i in range(0, len(content), 4):
        total = (total + struct.unpack(">L", content[i : i + 4])[0]) % MAX_UINT32
    return total == 0


def get_string(file: File) -> bytes:
    """Read a 16 bytes aligned, null terminated string."""
    filename = b""
    counter = 0
    while b"\x00" not in filename and counter < MAX_LINUX_PATH_LENGTH:
        filename += file.read(STRING_ALIGNMENT)
        counter += STRING_ALIGNMENT
    return filename.rstrip(b"\x00")


class FileHeader:
    addr: int
    next_filehdr: int
    spec_info: int
    fs_type: FSType
    executable: bool
    size: int
    checksum: int
    filename: bytes
    depth: int = -1
    parent: FileHeader | None = None
    start_offset: int
    end_offset: int
    file: File

    def __init__(self, addr: int, file: File):
        self.addr = addr
        fs_typeexec_next = struct.unpack(">L", file.read(4))[0]
        self.next_filehdr = fs_typeexec_next & ~0b1111
        self.fs_type = FSType(fs_typeexec_next & 0b0111)
        self.executable = fs_typeexec_next & 0b1000
        self.spec_info = struct.unpack(">I", file.read(4))[0]
        self.size = struct.unpack(">I", file.read(4))[0]
        self.checksum = struct.unpack(">I", file.read(4))[0]
        self.filename = get_string(file)
        self.start_offset = file.tell()
        self.file = file

    def valid_checksum(self) -> bool:
        current_position = self.file.tell()
        try:
            self.file.seek(self.addr, io.SEEK_SET)
            filename_len = len(self.filename)
            header_size = 16 + round_up(filename_len, 16)
            return valid_checksum(self.file.read(header_size))
        finally:
            self.file.seek(current_position, io.SEEK_SET)

    @property
    def content(self) -> bytes:
        """Returns the file content. Applicable to files and symlinks."""
        try:
            self.file.seek(self.start_offset, io.SEEK_SET)
            return self.file.read(self.size)
        finally:
            self.file.seek(-self.size, io.SEEK_CUR)

    @property
    def mode(self) -> int:
        """Permission mode.

        It is assumed to be world readable if executable bit is set,
        and world executable otherwise.  Handle mode for both block
        device and character devices too.
        """
        mode = WORLD_RWX if self.executable else WORLD_RW
        mode |= stat.S_IFBLK if self.fs_type == FSType.BLOCK_DEV else 0x0
        mode |= stat.S_IFCHR if self.fs_type == FSType.CHAR_DEV else 0x0
        return mode

    @property
    def dev(self) -> int:
        """Raw device number if block device or character device, zero otherwise."""
        if self.fs_type in [FSType.BLOCK_DEV, FSType.CHAR_DEV]:
            major = self.spec_info >> 16
            minor = self.spec_info & 0xFFFF
            return os.makedev(major, minor)
        return 0

    @property
    def path(self) -> Path:
        """Returns the full path of this file, up to the RomFS root."""
        current_node = self
        current_path = Path()
        while current_node is not None:
            current_path = Path(current_node.filename.decode("utf-8")).joinpath(
                current_path
            )
            current_node = current_node.parent
        return current_path

    def __repr__(self):
        return (
            f"FileHeader<next_filehdr:{self.next_filehdr}, type:{self.fs_type},"
            f" executable:{self.executable}, spec_info:{self.spec_info},"
            f" size:{self.size}, checksum:{self.checksum}, filename:{self.filename}>"
        )


class RomFSError(Exception):
    pass


class RomFSHeader:
    signature: bytes
    full_size: int
    checksum: int
    volume_name: bytes
    eof: int
    file: File
    end_offset: int
    inodes: dict[int, FileHeader]
    fs: FileSystem

    def __init__(
        self,
        file: File,
        fs: FileSystem,
    ):
        self.file = file
        self.file.seek(0, io.SEEK_END)
        self.eof = self.file.tell()
        self.file.seek(0, io.SEEK_SET)

        if self.eof < ROMFS_HEADER_SIZE:
            raise RomFSError("File too small to hold ROMFS")

        self.signature = self.file.read(8)
        self.full_size = struct.unpack(">I", self.file.read(4))[0]
        self.checksum = struct.unpack(">I", self.file.read(4))[0]
        self.volume_name = get_string(self.file)
        self.header_end_offset = self.file.tell()
        self.inodes = {}

        self.fs = fs

    def valid_checksum(self) -> bool:
        current_position = self.file.tell()
        try:
            self.file.seek(0, io.SEEK_SET)
            return valid_checksum(self.file.read(ROMFS_HEADER_SIZE))
        finally:
            self.file.seek(current_position, io.SEEK_SET)

    def validate(self):
        if self.signature != ROMFS_SIGNATURE:
            raise RomFSError("Invalid RomFS signature")
        if self.full_size > self.eof:
            raise RomFSError("ROMFS size is greater than file size")
        if not self.valid_checksum():
            raise RomFSError("Invalid checksum")

    def is_valid_addr(self, addr):
        """Validate that an inode address is valid.

        Inodes addresses must be 16 bytes aligned and placed within
        the RomFS on file.
        """
        return (self.header_end_offset <= addr <= self.eof) and (addr % 16 == 0)

    def is_recursive(self, addr) -> bool:
        return addr in self.inodes

    def recursive_walk(self, addr: int, parent: FileHeader | None = None):
        while self.is_valid_addr(addr) is True:
            addr = self.walk_dir(addr, parent)

    def walk_dir(self, addr: int, parent: FileHeader | None = None):
        self.file.seek(addr, io.SEEK_SET)
        file_header = FileHeader(addr, self.file)
        file_header.parent = parent

        if not file_header.valid_checksum():
            raise RomFSError(f"Invalid file CRC at addr {addr:0x}.")

        logger.debug("walking dir", addr=addr, file=file_header)

        if file_header.filename not in [b".", b".."]:
            if (
                file_header.fs_type == FSType.DIRECTORY
                and file_header.spec_info != 0x0
                and not self.is_recursive(addr)
            ):
                self.inodes[addr] = file_header
                self.recursive_walk(file_header.spec_info, file_header)
            self.inodes[addr] = file_header
        return file_header.next_filehdr

    def create_symlink(self, output_path: Path, inode: FileHeader):
        target_path = Path(inode.content.decode("utf-8"))
        self.fs.create_symlink(src=target_path, dst=output_path)

    def create_hardlink(self, output_path: Path, inode: FileHeader):
        if inode.spec_info in self.inodes:
            target_path = self.inodes[inode.spec_info].path
            self.fs.create_hardlink(dst=output_path, src=target_path)
        else:
            logger.warning("Invalid hard link target", inode_key=inode.spec_info)

    def create_inode(self, inode: FileHeader):
        output_path = inode.path
        logger.info("dumping inode", inode=inode, output_path=str(output_path))

        if inode.fs_type == FSType.HARD_LINK:
            self.create_hardlink(output_path, inode)
        elif inode.fs_type == FSType.SYMLINK:
            self.create_symlink(output_path, inode)
        elif inode.fs_type == FSType.DIRECTORY:
            self.fs.mkdir(output_path, mode=inode.mode, exist_ok=True)
        elif inode.fs_type == FSType.FILE:
            self.fs.write_bytes(output_path, inode.content)
        elif inode.fs_type in [FSType.BLOCK_DEV, FSType.CHAR_DEV]:
            self.fs.mknod(output_path, mode=inode.mode, device=inode.dev)
        elif inode.fs_type == FSType.FIFO:
            self.fs.mkfifo(output_path, mode=inode.mode)

    def dump_fs(self):
        def inodes(*inode_types):
            return sorted(
                (v for v in self.inodes.values() if v.fs_type in inode_types),
                key=lambda inode: inode.path,
            )

        # order of file object creation is important
        sorted_inodes = (
            inodes(FSType.FILE, FSType.DIRECTORY, FSType.FIFO, FSType.SOCKET)
            + inodes(FSType.BLOCK_DEV, FSType.CHAR_DEV)
            + inodes(FSType.SYMLINK, FSType.HARD_LINK)
        )

        for inode in sorted_inodes:
            self.create_inode(inode)

    def __str__(self):
        return f"signature: {self.signature}\nfull_size: {self.full_size}\nchecksum: {self.checksum}\nvolume_name: {self.volume_name}"


class RomfsExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as f:
            header = RomFSHeader(f, fs)
            header.validate()
            header.recursive_walk(header.header_end_offset, None)
            header.dump_fs()
            return ExtractResult(reports=fs.problems)


class RomFSFSHandler(StructHandler):
    NAME = "romfs"

    PATTERNS = [
        # '-rom1fs-'
        HexString("2D 72 6F 6D 31 66 73 2d")
    ]

    C_DEFINITIONS = r"""
        struct romfs_header {
            char magic[8];
            uint32 full_size;
            uint32 checksum;
        }
    """
    HEADER_STRUCT = "romfs_header"
    EXTRACTOR = RomfsExtractor()

    DOC = HandlerDoc(
        name="RomFS",
        description="RomFS is a simple, space-efficient, read-only file system format designed for embedded systems. It features 16-byte alignment, minimal metadata overhead, and supports basic file types like directories, files, symlinks, and devices.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="RomFS Documentation",
                url="https://www.kernel.org/doc/html/latest/filesystems/romfs.html",
            ),
            Reference(
                title="RomFS Wikipedia",
                url="https://en.wikipedia.org/wiki/Romfs",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        if not valid_checksum(file.read(512)):
            raise InvalidInputFormat("Invalid RomFS checksum.")

        file.seek(-512, io.SEEK_CUR)

        # Every multi byte value must be in big endian order.
        header = self.parse_header(file, Endian.BIG)

        # The zero terminated name of the volume, padded to 16 byte boundary.
        get_string(file)

        # seek filesystem size (number of accessible bytes in this fs)
        # from the actual end of the header
        file.seek(header.full_size, io.SEEK_CUR)

        # Another thing to note is that romfs works on file headers and data
        # aligned to 16 byte boundaries, but most hardware devices and the block
        # device drivers are unable to cope with smaller than block-sized data.
        # To overcome this limitation, the whole size of the file system must be
        # padded to an 1024 byte boundary.
        read_until_past(file, b"\x00")

        return ValidChunk(
            start_offset=start_offset,
            end_offset=file.tell(),
        )



from __future__ import annotations

import base64
import hashlib
import stat
import traceback
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any

if TYPE_CHECKING:
    from collections.abc import Iterable

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    computed_field,
    field_serializer,
    field_validator,
)


def ensure_bytes(value: Any) -> bytes:
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode()
    if value is None:
        return b""
    raise ValueError(f"Unsupported type in ensure_bytes: {type(value)}")


class Report(BaseModel):
    """A common base class for different reports. This will enable easy pydantic configuration of all models from a single point in the future if desired."""

    @computed_field
    @property
    def __typename__(self) -> str:
        return self.__class__.__name__


class Severity(Enum):
    """Represents possible problems encountered during execution."""

    ERROR = "ERROR"
    WARNING = "WARNING"


class ErrorReport(Report):
    severity: Severity


class UnknownError(ErrorReport):
    """Describes an exception raised during file processing."""

    severity: Severity = Severity.ERROR
    exception: str | Exception

    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )  # Necessary to support Exception type

    def model_post_init(self, _: Any) -> None:
        if isinstance(self.exception, Exception):
            self.exception = "".join(
                traceback.format_exception(
                    type(self.exception), self.exception, self.exception.__traceback__
                )
            )

    """Exceptions are also formatted at construct time."""


class CalculateChunkExceptionReport(UnknownError):
    """Describes an exception raised during calculate_chunk execution."""

    start_offset: int
    # Stored in `str` rather than `Handler`, because the pickle picks ups structs from `C_DEFINITIONS`
    handler: str


class CalculateMultiFileExceptionReport(UnknownError):
    """Describes an exception raised during calculate_chunk execution."""

    path: Path
    # Stored in `str` rather than `Handler`, because the pickle picks ups structs from `C_DEFINITIONS`
    handler: str


class ExtractCommandFailedReport(ErrorReport):
    """Describes an error when failed to run the extraction command."""

    severity: Severity = Severity.WARNING
    command: str
    stdout: Annotated[bytes, BeforeValidator(ensure_bytes)]
    stderr: Annotated[bytes, BeforeValidator(ensure_bytes)]
    exit_code: int

    # Use base64 to encode and decode bytes data in case there are non-standard characters
    @field_serializer("stdout", "stderr")
    def encode_bytes(self, v: bytes, _):
        return base64.b64encode(v).decode("ascii")

    @field_validator("stdout", "stderr", mode="before")
    @classmethod
    def decode_bytes(cls, v: Any):
        if isinstance(v, str):
            return base64.b64decode(v)
        return v


class OutputDirectoryExistsReport(ErrorReport):
    severity: Severity = Severity.ERROR
    path: Path


class ExtractorDependencyNotFoundReport(ErrorReport):
    """Describes an error when the dependency of an extractor doesn't exist."""

    severity: Severity = Severity.ERROR
    dependencies: list[str]


class ExtractorTimedOut(ErrorReport):
    """Describes an error when the extractor execution timed out."""

    severity: Severity = Severity.ERROR
    cmd: str
    timeout: float


class MaliciousSymlinkRemoved(ErrorReport):
    """Describes an error when malicious symlinks have been removed from disk."""

    severity: Severity = Severity.WARNING
    link: str
    target: str


class MultiFileCollisionReport(ErrorReport):
    """Describes an error when MultiFiles collide on the same file."""

    severity: Severity = Severity.ERROR
    paths: set[Path]
    handler: str


class StatReport(Report):
    path: Path
    size: int
    is_dir: bool
    is_file: bool
    is_link: bool
    link_target: Path | None

    @classmethod
    def from_path(cls, path: Path):
        st = path.lstat()
        mode = st.st_mode
        try:
            link_target = Path.readlink(path)
        except OSError:
            link_target = None

        return cls(
            path=path,
            size=st.st_size,
            is_dir=stat.S_ISDIR(mode),
            is_file=stat.S_ISREG(mode),
            is_link=stat.S_ISLNK(mode),
            link_target=link_target,
        )


class HashReport(Report):
    md5: str
    sha1: str
    sha256: str

    @classmethod
    def from_path(cls, path: Path):
        chunk_size = 1024 * 64
        md5 = hashlib.md5(usedforsecurity=False)
        sha1 = hashlib.sha1(usedforsecurity=False)
        sha256 = hashlib.sha256()

        with path.open("rb") as f:
            while chunk := f.read(chunk_size):
                md5.update(chunk)
                sha1.update(chunk)
                sha256.update(chunk)

        return cls(
            md5=md5.hexdigest(),
            sha1=sha1.hexdigest(),
            sha256=sha256.hexdigest(),
        )


class FileMagicReport(Report):
    magic: str
    mime_type: str


class RandomnessMeasurements(BaseModel):
    percentages: list[float]
    block_size: int
    mean: float

    @property
    def highest(self):
        return max(self.percentages)

    @property
    def lowest(self):
        return min(self.percentages)


class RandomnessReport(Report):
    shannon: RandomnessMeasurements
    chi_square: RandomnessMeasurements


class ChunkReport(Report):
    id: str
    handler_name: str
    start_offset: int
    end_offset: int
    size: int
    is_encrypted: bool
    extraction_reports: list[Report]

    @field_validator("extraction_reports", mode="before")
    @classmethod
    def validate_extraction_reports(cls, value: Any) -> list[Report]:
        return validate_report_list(value)


class UnknownChunkReport(Report):
    id: str
    start_offset: int
    end_offset: int
    size: int
    randomness: RandomnessReport | None

    @field_validator("randomness", mode="before")
    @classmethod
    def validate_randomness(cls, value: Any) -> RandomnessReport | None:
        if value is None:
            return None
        parsed = parse_report(value)
        if not isinstance(parsed, RandomnessReport):
            raise TypeError("Randomness must be a RandomnessReport.")
        return parsed


class CarveDirectoryReport(Report):
    carve_dir: Path


class MultiFileReport(Report):
    id: str
    handler_name: str
    name: str
    paths: list[Path]
    extraction_reports: list[Report]

    @field_validator("extraction_reports", mode="before")
    @classmethod
    def validate_extraction_reports(cls, value: Any) -> list[Report]:
        return validate_report_list(value)


class ExtractedFileDeletedReport(Report):
    path: Path
    handler_name: str


class ExtractionProblem(Report):
    """A non-fatal problem discovered during extraction.

    A report like this still means, that the extraction was successful,
    but there were problems that got resolved.
    The output is expected to be complete, with the exception of
    the reported path.

    Examples
    --------
    - duplicate entries for certain archive formats (tar, zip)
    - unsafe symlinks pointing outside of extraction directory

    """

    problem: str
    resolution: str
    path: str | None = None

    @property
    def log_msg(self):
        return f"{self.problem} {self.resolution}"

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path)


class PathTraversalProblem(ExtractionProblem):
    extraction_path: str

    def log_with(self, logger):
        logger.warning(
            self.log_msg,
            path=self.path,
            extraction_path=self.extraction_path,
        )


class LinkExtractionProblem(ExtractionProblem):
    link_path: str

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path, link_path=self.link_path)


class SpecialFileExtractionProblem(ExtractionProblem):
    mode: int
    device: int

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path, mode=self.mode, device=self.device)


BUILTIN_REPORT_TYPES: tuple[type[Report], ...] = (
    ErrorReport,
    UnknownError,
    CalculateChunkExceptionReport,
    CalculateMultiFileExceptionReport,
    ExtractCommandFailedReport,
    OutputDirectoryExistsReport,
    ExtractorDependencyNotFoundReport,
    ExtractedFileDeletedReport,
    ExtractorTimedOut,
    MaliciousSymlinkRemoved,
    MultiFileCollisionReport,
    StatReport,
    HashReport,
    FileMagicReport,
    RandomnessReport,
    ChunkReport,
    UnknownChunkReport,
    CarveDirectoryReport,
    MultiFileReport,
    ExtractionProblem,
    PathTraversalProblem,
    LinkExtractionProblem,
    SpecialFileExtractionProblem,
)

_REPORT_REGISTRY: dict[str, type[Report]] = {}


def register_report_type(report_type: type[Report]) -> None:
    typename = report_type.__name__
    existing = _REPORT_REGISTRY.get(typename)
    if existing is not None and existing is not report_type:
        raise ValueError(f"Report type name conflict: {typename}")
    _REPORT_REGISTRY[typename] = report_type


def register_report_types(report_types: Iterable[type[Report]]) -> None:
    for report_type in report_types:
        register_report_type(report_type)


def get_report_type(typename: str) -> type[Report] | None:
    return _REPORT_REGISTRY.get(typename)


def parse_report(report: Report | dict[str, Any]) -> Report:
    if isinstance(report, Report):
        return report
    if not isinstance(report, dict):
        raise TypeError("Report data must be a mapping or Report instance.")
    typename = report.get("__typename__")
    if not typename:
        raise ValueError("Report data is missing __typename__.")
    report_type = get_report_type(typename)
    if report_type is None:
        raise ValueError(f"Unknown report type: {typename}")
    return report_type.model_validate(report)


def validate_report_list(value: Any) -> list[Report]:
    if not isinstance(value, list):
        raise TypeError("Report list must be a list.")
    return [parse_report(item) for item in value]


register_report_types(BUILTIN_REPORT_TYPES)



END_OF_STREAM_MARKER = 0xFF_FF_FF_FF


class BitReader:
    def __init__(self, data: bytes, start: int = 0):
        self.data = data
        self.index = start  # index in the data (byte-wise)
        self.bb = 0  # bit buffer
        self.bc = 0  # number of bits remaining in the buffer

    def get_bit(self) -> int:
        """Return the next bit from the data stream (0 or 1). When the bit buffer is empty, loads the next byte."""
        if self.bc == 0:
            if self.index >= len(self.data):
                raise ValueError("Unexpected end of data while reading bit")
            self.bb = self.data[self.index]
            self.index += 1
            self.bc = 8
        self.bc -= 1
        return (self.bb >> self.bc) & 1

    def read_byte(self) -> int:
        """Read and return the next full byte from the data stream. This does not take into account any bits already buffered."""
        if self.index >= len(self.data):
            raise ValueError("Unexpected end of data while reading byte")
        b = self.data[self.index]
        self.index += 1
        return b


class UCLDecompressor:
    def __init__(self):
        self._reader: BitReader = BitReader(b"")
        self._output: bytearray = bytearray()
        self._last_match_offset: int = -1
        self._match_offset: int = -1
        self._match_length: int = -1

    def _process_literal_run(self) -> None:
        """Process a run of literal bytes while the next bit is 1."""
        while self._reader.get_bit():
            self._output.append(self._reader.read_byte())

    def _decode_match_offset(self) -> int:
        """Decode the match offset value from the bit stream."""
        match_offset = 1
        while True:
            match_offset = (match_offset << 1) + self._reader.get_bit()
            if self._reader.get_bit() == 1:
                break
        return match_offset

    def _decode_match_length(self) -> int:
        """Decode the match length value from the bit stream."""
        # Get a two-bit base for the match length
        match_length = (self._reader.get_bit() << 1) + self._reader.get_bit()

        if match_length == 0:
            match_length += 1
            # Read extra bits until a terminating 1 is encountered
            while True:
                match_length = (match_length << 1) + self._reader.get_bit()
                if self._reader.get_bit() == 1:
                    break
            match_length += 2

        # If the match offset is large, add an extra byte to the length.
        if self._match_offset > 0xD00:
            match_length += 1

        return match_length

    def _copy_match_data(self) -> None:
        """Copy match data from the already decompressed output."""
        match_index = len(self._output) - self._match_offset
        if match_index < 0:
            raise ValueError("Invalid match offset")

        # Copy one byte unconditionally
        self._output.append(self._output[match_index])
        match_index += 1

        # Then copy match_length bytes (the regions may overlap)
        for _ in range(self._match_length):
            self._output.append(self._output[match_index])
            match_index += 1

    def decompress(self, compressed: bytes) -> bytes:
        """UCL decompression using NRV2B mode."""
        self._reader = BitReader(compressed, start=0)
        self._output = bytearray()
        self._last_match_offset = 1

        while True:
            self._process_literal_run()
            self._match_offset = self._decode_match_offset()

            if self._match_offset == 2:
                self._match_offset = self._last_match_offset
            else:
                # Read an extra byte to complete the offset.
                self._match_offset = (
                    self._match_offset - 3
                ) * 256 + self._reader.read_byte()
                if self._match_offset == END_OF_STREAM_MARKER:
                    break
                self._match_offset += 1
                self._last_match_offset = self._match_offset

            self._match_length = self._decode_match_length()
            self._copy_match_data()

        return bytes(self._output)



import io

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, convert_int64
from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)

logger = get_logger()

# magic (4 bytes) + VN (1 byte) + DS (1 byte)
HEADER_LEN = 4 + 1 + 1
# LZMA stream is 2 bytes aligned
LZMA_ALIGNMENT = 2


class LZipHandler(Handler):
    NAME = "lzip"

    PATTERNS = [HexString("4C 5A 49 50 01")]

    EXTRACTOR = Command(
        "lziprecover", "-k", "-D0", "-i", "{inpath}", "-o", "{outdir}/lz.uncompressed"
    )

    DOC = HandlerDoc(
        name="Lzip",
        description="Lzip is a lossless compressed file format based on the LZMA algorithm. It features a simple header, CRC-checked integrity, and efficient compression for large files.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="Lzip File Format Documentation",
                url="https://www.nongnu.org/lzip/manual/lzip_manual.html",
            ),
            Reference(
                title="Lzip Wikipedia",
                url="https://en.wikipedia.org/wiki/Lzip",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(HEADER_LEN, io.SEEK_CUR)
        # quite the naive idea but it works
        # the idea is to read 8 bytes uint64 every 2 bytes alignment
        # until we end up reading the Member Size field which corresponds
        # to "the total size of the member, including header and trailer".
        # We either find it or reach EOF, which will be caught by finder.

        while True:
            file.seek(LZMA_ALIGNMENT, io.SEEK_CUR)
            member_size = convert_int64(file.read(8), Endian.LITTLE)
            if member_size == (file.tell() - start_offset):
                end_offset = file.tell()
                break
            file.seek(-8, io.SEEK_CUR)

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)
